"""
Common utilities for payload generation
"""
import os
import shutil
import grp
import pwd
from typing import Literal

from snr.core.core import context, options
from snr.core.util import common_utils, programs

__all__ = (
    "clean_on_success", "clean_and_exit"
)


def is_user_disk_admin() -> bool:
    """Is the user a disk administrator
    Being a disk administrator means either being in `disk` group, or being root

    Returns:
        Whatever the current user is a disk admin or not
    """
    if os.getuid() == 0:
        return True
    username = pwd.getpwuid(os.getuid()).pw_name
    try:
        for group in grp.getgrall():
            if username in group.gr_mem and group.gr_name == "disk":
                return True
        return False
    except KeyError:
        return False


def bind_required_rootfs_dirs(ctx: context.Context) -> None:
    """Bind required directories for a chroot environment

    Args:
        ctx: Context
    """
    for node, check_measure in [("dev", "dev/random"), ("sys", "sys/kernel"), ("proc", "proc/self")]:
        if os.path.exists(os.path.join(ctx.root_directory, check_measure)):
            common_utils.print_debug(f"/{node} is already mounted")
            continue
        common_utils.print_debug(f"Mounting /{node}")
        os.makedirs(os.path.join(ctx.root_directory, node), exist_ok=True)
        mount = programs.Mount(sudo=True)

        errorcode = mount.invoke_and_wait(None,
                                          "-B",
                                          f"/{node}",
                                          os.path.join(ctx.root_directory, node))
        if errorcode != 0:
            assert mount.stdout is not None
            common_utils.print_debug(
                f"Command's output: {mount.stdout.read()}")
            clean_and_exit(
                ctx, "Binding required rootfs directories for a chroot failed", True, True)


def unbind_required_rootfs_dirs(ctx: context.Context) -> None:
    """Unbind required directories for a chroot environment, as well as boot/efi

    Args:
        ctx: Context
    """
    for node in ["sys", "dev", "proc", "boot/efi"]:
        common_utils.print_debug(f"Unmounting /{node}")
        umount = programs.Umount(sudo=True,
                                 stdout=programs.PIPE,
                                 stderr=programs.STDOUT)

        errorcode = umount.invoke_and_wait(None,
                                           os.path.join(
                                               ctx.root_directory, node),
                                           options={})
        if errorcode != 0:
            assert umount.stdout is not None
            output = umount.stdout.read()
            if "not mounted" not in output:
                common_utils.print_debug(
                    f"Command's output: {output}")


def clean_on_success(ctx: context.Context,
                     unmount_loop: bool = False,
                     clear_root_directory: bool = False) -> Literal[True]:
    """Clean up after success.

    Args:
        ctx: Context
        unmount_loop: Whether or not to unmount the loop device. Defaults to False
        clear_root_directory: Whether or not to clear root_directory. Defaults to False

    Returns:
        Always returns True
    """
    if not ctx.is_device and unmount_loop:
        common_utils.print_debug("Unmounting rootfs filesystem")
        umount = programs.Umount(
            sudo=True,
            stdout=programs.DEVNULL,
            stderr=programs.DEVNULL)
        umount.invoke_and_wait(None, ctx.root_directory)
        common_utils.print_debug("Deleting loop")
        losetup = programs.Losetup(
            sudo=not is_user_disk_admin(),
            stdout=programs.DEVNULL,
            stderr=programs.DEVNULL)
        losetup.invoke_and_wait(None, ctx.device_name,
                                options={
                                    "detach": None
                                })
    if clear_root_directory and ctx.level >= 3:
        unbind_required_rootfs_dirs(ctx)
        common_utils.print_debug(
            "Changing file ownerships back to user for deletion")
        chroot = programs.Chown(sudo=True)
        chroot.invoke_and_wait(None, f"{os.getuid()}:{os.getgid()}", ctx.root_directory, options={
            "recursive": None,
        })
        common_utils.print_debug("Cleaning up root directory")
        shutil.rmtree(ctx.root_directory, ignore_errors=True)
    return True


def clean_and_exit(ctx: context.Context, message: str,
                   unmount_loop: bool = False, clear_root_directory: bool = False) -> Literal[False]:
    """Clean and exit with error

    Args:
        ctx: Context 
        message: error message to pass to clean_on_success
        unmount_loop: if True unmount the loop
        clear_root_directory: Whether or not to clear root_directory. Defaults to False

    Returns:
        Always return False
    """
    if not options.debug:
        clean_on_success(ctx, unmount_loop, clear_root_directory)
    else:
        # Still for safety reasons it's best to do this
        unbind_required_rootfs_dirs(ctx)
    common_utils.print_fatal(message)
    return False



"""
Utility functions to log output to the screen and the log file,
@note Do not use the print_sys function in anywhere with the exception of the snr cli code.
"""
import io
import os.path
import re
import sys
import time
from typing import TextIO

import rich.console
import rich.pretty

from snr.core.core import common_paths, options

__all__ = (
    "SYS", "DEBUG",
    "INFO", "OK",
    "WARNING", "ERROR",
    "carriage_return", "clear_screen",
    "print_sys", "print_debug",
    "print_info", "print_ok",
    "print_warning", "print_error",
    "print_fatal"
)

LOG_FILE_PATH: str = os.path.join(common_paths.STATE_PATH, "main.log")
ERROR_LOG_FILE_PATH: str = os.path.join(common_paths.STATE_PATH, "errors.log")

LOG_FILE_STREAM: io.TextIOWrapper
ERROR_LOG_FILE_STREAM: io.TextIOWrapper

# System message symbol
SYS = "-->"
# Debug message symbol
DEBUG = "[[magenta].[/magenta]]"
# Informational message symbol
INFO = "[[blue]![/blue]]"
# Successful message symbol
OK = "[[green]+[/green]]"
# Warning message symbol
WARNING = "[[yellow]*[/yellow]]"
# Error message symbol
ERROR = "[[red]-[/red]]"

_ANSI_ESCAPE = re.compile(r'\x1b\[[0-9;]*m')


def carriage_return() -> None:
    """Return to the previous line.
    """
    sys.stdout.write('\r')


def clear_screen() -> None:
    """Clears the screen.
    """
    print("\x1b[H\x1b[2J\x1b[3J")


def _stringize(obj: object) -> str:
    if isinstance(obj, rich.pretty.Pretty):
        return _ANSI_ESCAPE.sub('', rich.pretty.pretty_repr(obj._object).strip())
    return str(obj)


def _print_raw(*args: object, sep: str = ' ',
               end: str = '\n', file: TextIO = sys.stderr, flush: bool = False) -> None:

    s = sep.join(map(_stringize, args))
    if file == sys.stderr:
        console = rich.console.Console()
    else:
        console = rich.console.Console(file=file)
    if s.startswith(INFO):
        if not options.quiet:
            console.print(s, sep=sep, end=end)
            if flush:
                console.file.flush()
    elif s.startswith(DEBUG):
        if options.verbose:
            console.print(s, sep=sep, end=end)
            if flush:
                console.file.flush()
    else:
        console.print(s, sep=sep, end=end)
        if flush:
            console.file.flush()
    LOG_FILE_STREAM.write(
        f"[{int(time.time())}] {_ANSI_ESCAPE.sub('', s)}")
    LOG_FILE_STREAM.flush()
    if str(args[0]).startswith(ERROR):
        ERROR_LOG_FILE_STREAM.write(
            f"[{int(time.time())}] {_ANSI_ESCAPE.sub('', s)}")
        ERROR_LOG_FILE_STREAM.flush()


def print_sys(*args: object, sep: str = ' ',
              end: str = '\n', file: TextIO = sys.stderr, flush: bool = False) -> None:
    """Print a message to the log file with the system symbol.

    Args:
        args: The arguments to print.
        sep: The separator between the arguments.
        end: The end of the message.
        file: The file to print to.
        flush: Whether to flush the output.
    """
    _print_raw(SYS, *args, sep=sep, end=end, file=file, flush=flush)


def print_debug(*args: object, sep: str = ' ',
                end: str = '\n', file: TextIO = sys.stderr, flush: bool = False) -> None:
    """Print a message to the log file with the debug symbol, but only if on verbose mode.

    Args:
        args: The arguments to print.
        sep: The separator between the arguments.
        end: The end of the message.
        file: The file to print to.
        flush: Whether to flush the output.
    """
    _print_raw(DEBUG, *args, sep=sep, end=end, file=file, flush=flush)


def print_info(*args: object, sep: str = ' ',
               end: str = '\n', file: TextIO = sys.stderr, flush: bool = False) -> None:
    """Print a message to the log file with the info symbol, but only if on quiet mode.

    Args:
        args: The arguments to print.
        sep: The separator between the arguments.
        end: The end of the message.
        file: The file to print to.
        flush: Whether to flush the output.
    """
    _print_raw(INFO, *args, sep=sep, end=end, file=file, flush=flush)


def print_ok(*args: object, sep: str = ' ',
             end: str = '\n', file: TextIO = sys.stderr, flush: bool = False) -> None:
    """Print a message to the log file with the ok symbol.

    Args:
        args: The arguments to print.
        sep: The separator between the arguments.
        end: The end of the message.
        file: The file to print to.
        flush: Whether to flush the output.
    """
    _print_raw(OK, *args, sep=sep, end=end, file=file, flush=flush)


def print_warning(*args: object, sep: str = ' ',
                  end: str = '\n', file: TextIO = sys.stderr, flush: bool = False) -> None:
    """Print a message to the log file with the warning symbol.

    Args:
        args: The arguments to print.
        sep: The separator between the arguments.
        end: The end of the message.
        file: The file to print to.
        flush: Whether to flush the output.
    """
    _print_raw(WARNING, *args, sep=sep, end=end, file=file, flush=flush)


def print_error(*args: object, sep: str = ' ',
                end: str = '\n', file: TextIO = sys.stderr, flush: bool = False) -> None:
    """Print a message to the log file with the error symbol.

    Args:
        args: The arguments to print.
        sep: The separator between the arguments.
        end: The end of the message.
        file: The file to print to.
        flush: Whether to flush the output.
    """
    _print_raw(ERROR, *args, sep=sep, end=end, file=file, flush=flush)


def print_fatal(*args: object, sep: str = ' ',
                end: str = '\n', file: TextIO = sys.stderr, flush: bool = False) -> None:
    """Print a message to the log file with the error symbol, exiting afterwards.
    Exit code is taken from snr.core.options.default_exit_code

    Args:
        args: The arguments to print.
        sep: The separator between the arguments.
        end: The end of the message.
        file: The file to print to.
        flush: Whether to flush the output.
    """
    _print_raw(ERROR, *args, sep=sep, end=end, file=file, flush=flush)
    if not options.debug:
        raise SystemExit(options.default_exit_code)
    raise RuntimeError(sep.join(map(str, args)))


def _try_open_log_files() -> None:
    global LOG_FILE_STREAM  # pylint: disable=global-statement
    global ERROR_LOG_FILE_STREAM  # pylint: disable=global-statement

    if hasattr(_try_open_log_files, "has_already_done"):
        return

    try:
        for file in (LOG_FILE_PATH, ERROR_LOG_FILE_PATH):
            if not os.path.exists(LOG_FILE_PATH):
                with open(file, "w", encoding="utf-8") as _:
                    pass
        LOG_FILE_STREAM = open(LOG_FILE_PATH,  # pylint: disable=consider-using-with
                               "a", encoding="utf-8")
        ERROR_LOG_FILE_STREAM = open(ERROR_LOG_FILE_PATH,  # pylint: disable=consider-using-with
                                     "a", encoding="utf-8")
    except OSError as exc:
        # We cannot fail because the logger cannot open a file
        # Assume a readonly filesystem and use an IO wrapper
        LOG_FILE_STREAM = io.TextIOWrapper(io.BytesIO())
        ERROR_LOG_FILE_STREAM = io.TextIOWrapper(io.BytesIO())

        print_sys(
            f"Failed to open log file: Using in-memory buffer instead ({exc})")
    setattr(_try_open_log_files, "has_already_done", True)
    print_debug("Logger started")


_try_open_log_files()



"""
Standard snr shell atexit callbacks
"""
import atexit
import os

from snr.core.util import common_utils

_old_root = os.getcwd()

__all__ = (
    "restore_cwd",
    "register_atexit_callbacks", "unregister_atexit_callbacks"
)


def restore_cwd() -> None:
    """
    Restore CWD to what it was once this file was imported
    """
    common_utils.print_info("Restoring CWD...")
    os.chdir(_old_root)


def register_atexit_callbacks() -> None:
    """
    Register all atexit callbacks
    """
    atexit.register(restore_cwd)


def unregister_atexit_callbacks() -> None:
    """
    Unregister all atexit callbacks
    """
    atexit.unregister(restore_cwd)



"""
Satisfy the reported payload dependencies
"""

from snr.core.core import context, options
from snr.core.payload_generation import common
from snr.core.util import chroot_programs, common_utils

__all__ = (
    "ensure_dependencies",
)


def ensure_dependencies(ctx: context.Context) -> bool:
    """Install dependencies of the payload

    Args:
        ctx: Context

    Returns:
        Whatever operation was successful or not
    """
    assert options.payload_module is not None
    if len(options.payload_module.payload.DEPENDENCIES) != 0:  # pylint: disable=no-member
        dependencies = options.payload_module.payload.DEPENDENCIES  # pylint: disable=no-member
        common_utils.print_debug(
            "Installing dependencies of the payload:" +
            ' '.join(dependencies))
        apt_get = chroot_programs.apt_get_factory(ctx)(
            stdout=chroot_programs.PIPE,
            stderr=chroot_programs.STDOUT,
        )

        errorcode = apt_get.invoke_and_wait(None,
                                            *dependencies,
                                            options={
                                                "quiet": "2"
                                            })
        if errorcode != 0:
            assert apt_get.stdout is not None
            common_utils.print_debug(
                f"Command's output: {apt_get.stdout.read()}")
            return common.clean_and_exit(ctx, "Installing dependencies of payload failed")
    else:
        common_utils.print_debug("Payload has no dependencies, nothing to do")
    return True



"""
Drop to a shell as root
"""

from snr.core.payload.payload import Payload, Context


class ShellPayload(Payload):
    AUTHORS = ("GlobularOne",)

    def generate(self, _: Context) -> int:
        # Do nothing, really nothing
        return 0


payload = ShellPayload()



"""
Payload class base
"""

import os.path
import shutil
from typing import Mapping, Optional

from snr.cli import variables
# This is an exception to the general rule, it is best for payloads to
# get their Context class from payload, so they can use that name for `snr.core.payload.context`
from snr.core.core.context import Context
from snr.core.core.variable_manager import VariableType
from snr.core.util import at_formatter, common_utils
from snr.core.util.payloads import autorun

__all__ = (
    "Payload", "Context"
)


class Payload:
    """Base payload class. Contains utilities for payload generation as well.

    Attributes:

        AUTHORS:
            Tuple of authors. In format "Pseudonym (Real name) <Email>".
            In place of pseudonym, real name can be used and all parts but the name is optional

        LICENSE: 
            License SPDX identifier, not specifying one means using 
            the project's license (GPLv3)

        INPUTS: 
            A tuple of input variables. More specifically, it's a tuple of tuples,
            each tuple is for one variable. Order is, variable name, 
            default value, size, documentation An example would be:
            (
                ("foo", "bar", 3, "Foo or bar"),
                ("spam", [], 12, "12 Spams")
            )

        DEPENDENCIES: Deb packages as dependencies for the payload to work
    """
    AUTHORS: tuple[Optional[str], ...] = ()
    LICENSE: str = "gpl-3.0"
    INPUTS: tuple[tuple[str, VariableType, int, str], ...] = ()
    DEPENDENCIES: tuple[str, ...] = ()

    _autorun: autorun.Autorun | None = None

    def load(self) -> int:
        """Payload load hook

        Returns:
            status code (0 for success, anything else for failure)
        """
        for inp in self.INPUTS:
            variables.global_vars.set_variable(
                inp[0], inp[1], inp[2], inp[3], True)
        return 0

    def unload(self) -> int:
        """Payload unload hook

        Returns:
            status code (0 for success, anything else for failure)
        """
        for inp in self.INPUTS:
            variables.global_vars.del_variable(inp[0])
        return 0

    def generate(self, ctx: Context) -> int:
        """Payload generate method

        Args:
            ctx: Payload generation context

        Returns:
            status code (0 for success, anything else for failure)
        """
        raise NotImplementedError

    def copy_root_to_root(self, ctx: Context, module_path: str,
                          src: str, dest: str,
                          follow_symlinks: bool = True) -> str:
        """Copy a file from payload's package root to host filesystem's root

        Args:
            ctx: Payload generation context
            module_path: Path to the payload's __init__.py module
            src: copy operation source
            dest: copy operation destination
            follow_symlinks Follow symlinks instead of recreating them. Defaults to True

        Returns:
            Same as shutil.copyfile
        """
        return shutil.copyfile(os.path.join(os.path.dirname(module_path), src), os.path.join(
            ctx.root_directory, dest), follow_symlinks=follow_symlinks)

    def format_payload_and_write(self, ctx: Context, data: Mapping[str, VariableType],
                                 local_payload_path: str = "payload.py",
                                 host_payload_path: str = "root/payload.py") -> None:
        """Format a payload using AtFormatter and write it to the host filesystem

        Args:
            ctx: Payload generation context
            data: Variable, Value dictionary
            local_payload_path: Local path to payload
            host_payload_path: Path inside the host filesystem to write the result to
        """
        with open(local_payload_path, encoding="utf-8") as stream:
            payload_data = stream.read()
        with common_utils.rootfs_open(ctx, host_payload_path, "w") as stream:
            stream.write(at_formatter.AtFormatter(
                data).format_str(payload_data))

    def get_self_variables(self) -> dict[str, VariableType]:
        """Obtain all variables related to this payload.
        Variables related to this payload are discovered using the payload's INPUTS

        Returns:
            All variables related to this payload
        """
        result = {}
        for inp in self.INPUTS:
            result[inp[0]] = variables.global_vars.get_variable(inp[0])
        return result

    def add_autorun(self, ctx: Context, executable: str = "root/payload.py") -> None:
        """Add an autorun service for executable.
        In case of several calls, order will be preserved

        Args:
            ctx: Payload generation context
            executable: Executable to add service for (it must exist on host filesystem)
        """
        if self._autorun is None:
            self._autorun = autorun.Autorun(ctx)
        self._autorun.add_executable(executable).write()



"""
Module containing a class providing a more clean interface to work with systemd unit files
"""
import configparser
import os
import pprint
import warnings

from snr.core.core import context
from snr.core.util import common_utils

__all__ = (
    "SYSTEMD_SYSTEM_PATH", "SystemdConfigFileBase",
    "SystemdUnit"
)


def SYSTEMD_SYSTEM_PATH(ctx: context.Context) -> str:  # pylint: disable=invalid-name
    """Return the path to systemd's system directory
    
    Args:
        ctx: A dictionary containing context information
    
    Returns:
        The path to systemd's system directory
    """
    return os.path.join(ctx.root_directory, "usr", "lib", "systemd", "system")


_ValidOptionValueType = str | int | bool
# Type for a system section
SystemdSectionType = dict[str, _ValidOptionValueType]


class _SystemdConfigParser(configparser.ConfigParser):
    def optionxform(self, optionstr: str) -> str:
        return optionstr


class SystemdConfigFileBase:
    """Provide basic systemd unit file parsing, is not usually used directly.
    """
    _base_sections: tuple[str, ...] = ()
    _extra_sections: tuple[str, ...] = ()
    root: str = ""
    path: str = ""
    suffix: str = ""
    basename: str = ""

    def __init_subclass__(cls, root: str, base_sections: tuple[str, ...] = ()) -> None:
        cls.root = root
        cls._base_sections = tuple(base_sections)

    def read(self) -> None:
        """Read and parse configuration file

        Raises:
            FileNotFoundError: If the config file does not exist
        """
        parser = _SystemdConfigParser()
        if len(parser.read(self.path)) != 1:
            raise FileNotFoundError(
                f"No such file or directory: {self.path}")
        section_names = parser.sections()
        for section_name in section_names:
            section_name = section_name.replace("-", "_")
            if section_name in self._extra_sections or section_name in self._base_sections:
                setattr(self, f"{section_name}_section",
                        {**parser[section_name]})
            else:
                warnings.warn(
                    f"SystemdUnit: Unrecognized section '{section_name}'")

    def _write(self) -> None:
        """Write the unit file
        """
        sections = {}
        for section_name in self._base_sections:
            sections[section_name.replace(
                "_", "-")] = getattr(self, f"{section_name}_section")
        for section_name in self._extra_sections:
            sections[section_name.replace(
                "_", "-")] = getattr(self, f"{section_name}_section")
        common_utils.print_debug(
            f"Writing unit file {self.basename} with data:\n{pprint.pformat(sections)}")
        parser = _SystemdConfigParser()
        parser.read_dict(sections)
        with open(self.path, "w", encoding="utf-8") as stream:
            parser.write(stream, False)

    def write(self) -> None:
        """Write unit the file
        """
        self._write()


class SystemdUnit(SystemdConfigFileBase,
                  root="/nonexistent",
                  base_sections=("Unit", "Install")):
    """Provide a more clean interface to work with systemd unit files
    """
    Unit_section: SystemdSectionType = {}
    Install_section: SystemdSectionType = {}
    _context: context.Context

    def __init__(self, ctx: context.Context, name: str):
        for extra_section in self._extra_sections:
            setattr(self, f"{extra_section}_section", {})
        for base_section in self._base_sections:
            setattr(self, f"{base_section}_section", {})
        self.root = SYSTEMD_SYSTEM_PATH(ctx)
        self.path = os.path.join(self.root, name + self.suffix)
        self.basename = os.path.basename(self.path)
        self._context = ctx

    def __init_subclass__(cls, suffix: str, extra_sections: tuple[str, ...] = ()):
        cls.suffix = suffix
        cls._extra_sections = tuple(extra_sections)

    def write(self, make_wants_dir: bool = True, make_requires_dir: bool = True) -> None:
        """Write the config to disk
        
        Args:
            make_wants_dir: Whether or not to make the .wants directory.
            make_requires_dir: Whether or not to make the .requires directory
        """
        self._write()
        if make_wants_dir:
            os.mkdir(os.path.join(self.root, self.basename) + ".wants")
        if make_requires_dir:
            os.mkdir(os.path.join(self.root, self.basename) + ".requires")



"""
Install GRUB for both UEFI and BIOS boot on the host
"""
import os
import os.path

from snr.core.core import context
from snr.core.payload_generation import common
from snr.core.util import chroot_programs, common_utils, programs

__all__ = (
    "install_grub",
)


def install_grub(ctx: context.Context) -> bool:
    """Install GRUB for both UEFI and BIOS boot on the target

    Args:
        ctx: Context

    Returns:
        Whatever operation was successful or not
    """
    boot_mount_info = (
        (ctx.construct_partition_path(2), "boot/efi"),
    )
    for src, dest in boot_mount_info:
        common_utils.print_debug(f"Mounting /{dest}")
        mount = programs.Mount(sudo=True)

        errorcode = mount.invoke_and_wait(None,
                                          src,
                                          os.path.join(ctx.root_directory, dest))
        if errorcode != 0:
            assert mount.stdout is not None
            common_utils.print_debug(
                f"Command's output: {mount.stdout.read()}")
            return common.clean_and_exit(ctx, "Preparing to install grub failed", True, True)

    common.bind_required_rootfs_dirs(ctx)
    common_utils.print_debug("Installing UEFI grub")
    grub_install = chroot_programs.grub_install_factory(ctx)(
        stdout=chroot_programs.PIPE,
        stderr=chroot_programs.STDOUT,
        sudo=True)
    errorcode = grub_install.invoke_and_wait(None,
                                             ctx.device_name,
                                             options={
                                                 "uefi-secure-boot": None,
                                                 "removable": None
                                             })
    if errorcode != 0:
        assert grub_install.stdout is not None
        common_utils.print_debug(
            f"Command's output: {grub_install.stdout.read()}")
        common.unbind_required_rootfs_dirs(ctx)
        return common.clean_and_exit(ctx, "Installing grub failed", True, True)

    common_utils.print_debug("Install BIOS grub")
    grub_install = chroot_programs.grub_install_factory(ctx)(
        stdout=chroot_programs.PIPE,
        stderr=chroot_programs.STDOUT,
        sudo=True)
    errorcode = grub_install.invoke_and_wait(None,
                                             ctx.device_name,
                                             options={"target": "i386-pc"})
    if errorcode != 0:
        common.unbind_required_rootfs_dirs(ctx)
        assert grub_install.stdout is not None
        common_utils.print_debug(
            f"Command's output: {grub_install.stdout.read()}")
        return common.clean_and_exit(ctx, "Installing grub failed", True, True)
    common.unbind_required_rootfs_dirs(ctx)
    return True



"""
Format the partitions on the host
"""
import os

from snr.core.core import context
from snr.core.payload_generation import common
from snr.core.util import common_utils, programs

__all__ = (
    "format_host",
)


def format_host(ctx: context.Context) -> bool:
    """Format partitions on the host

    Args:
        ctx: Context

    Returns:
        Whatever operation was successful or not
    """
    common_utils.print_debug("Formatting partitions")
    for index, tool, flags in (
        ("2", "mkfs.vfat", ("-F32",)),
        ("3", "mkfs.ext4", ("-F", "-L", "Rootfs",
         f"-Eroot_owner={os.getuid()}:{os.getgid()}"))
    ):
        part = ctx.construct_partition_path(index)
        tool_wrapper = programs.program_wrapper_factory(tool)(
            sudo=not common.is_user_disk_admin(),
            stdout=programs.PIPE,
            stderr=programs.STDOUT)
        errorcode = tool_wrapper.invoke_and_wait(None,
                                                 *flags,
                                                 part)
        if errorcode != 0:
            assert tool_wrapper.stdout is not None
            common_utils.print_debug(f"Command's output: {tool_wrapper.stdout.read()}")
            return common.clean_and_exit(ctx, "Formatting failed", True)
    return True



#!/usr/bin/python3
"""
"""
import shutil
import time

from snr.core.util import common_utils
from snr.core.payload import entry_point

TERM_COLOR = "\033[41;30m"
FLAG = r"""\
                               ______
                            .-"      "-.
                           /            \
               _          |              |          _
              ( \         |,  .-.  .-.  ,|         / )
               > "=._     | )(__/  \__)( |     _.=" <
              (_/"=._"=._ |/     /\     \| _.="_.="\_)
                     "=._ (_     ^^     _)"_.="
                         "=\__|IIIIII|__/="
                        _.="| \IIIIII/ |"=._
              _     _.="_.="\          /"=._"=._     _
             ( \_.="_.="     `--------`     "=._"=._/ )
              > _.="                            "=._ <
             (_/                                    \_)
"""


@entry_point.entry_point
def main() -> None:
    common_utils.clear_screen()
    print(TERM_COLOR, end="")
    terminal_size = shutil.get_terminal_size()
    print(" " * (terminal_size.columns * terminal_size.lines), end="")
    print(FLAG)
    while True:
        time.sleep(10)


if __name__ == "__main__":
    main()



"""Snr shell commands lexer
"""
import pygments.lexer
import pygments.token

__all__ = (
    "SnrLexer",
)

CHECKSUM_ALGORITHMS = (
    "blake2b", "md5", "sha1",
    "sha224", "sha256", "sha384", "sha512"
)


class SnrLexer(pygments.lexer.RegexLexer):  # type: ignore
    """Lexer for snr shell commands"""
    name = 'SnrLexer'

    tokens = {
        'root': [
            pygments.lexer.include('commands')
        ],
        'commands': [
            pygments.lexer.include('pwd'),
            pygments.lexer.include('chdir'),
            pygments.lexer.include('list'),
            pygments.lexer.include('read'),
            pygments.lexer.include('checksum'),
            pygments.lexer.include('unset'),
            pygments.lexer.include('set'),
            pygments.lexer.include('use'),
            pygments.lexer.include('generate'),
            pygments.lexer.include('clear'),
            pygments.lexer.include('echo'),
            pygments.lexer.include('exit'),
            pygments.lexer.include('help'),
            pygments.lexer.include('info'),
            pygments.lexer.include('pdb'),
            pygments.lexer.include('reload')
        ],
        'eoc': [
            (r'\s', pygments.token.Whitespace),
            (r'#.*$', pygments.token.Comment.Single),
        ],
        'basic': [
            (r'(?<!\\)\$\w+', pygments.token.Name.Variable),
            (r'\b\d+\b', pygments.token.Number),
            (r'\s', pygments.token.Whitespace),
            (r'(?<!\!).+', pygments.token.Text),
        ],
        'pwd': [
            (r'(\s*)(pwd)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
             ), 'eoc'),
        ],
        'chdir': [
            (r'(\s*)(chdir)(\s+)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
                 pygments.token.Whitespace,
             ), 'basic'),
        ],
        'list': [
            (r'(\s*)(list)(\s+)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
                 pygments.token.Whitespace,
             ), 'basic'),
            (r'(\s*)(list)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
             ), 'eoc'),
        ],
        'read': [
            (r'(\s*)(read)(\s+)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
                 pygments.token.Whitespace,
             ), 'basic'),
        ],
        'checksum': [
            (r'(\s*)(checksum)(\s+)' + f"({'|'.join(CHECKSUM_ALGORITHMS)})" + r'(\s+)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
                 pygments.token.Whitespace,
                 pygments.token.Keyword.Constant,
                 pygments.token.Whitespace,
             ), 'basic'),
        ],
        'unset': [
            (r'(\s*)(unset)(\s+)(\w+)(\s*)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
                 pygments.token.Whitespace,
                 pygments.token.Name.Variable,
                 pygments.token.Whitespace
             ), 'eoc'),
        ],
        'set': [
            # Set assign to command output
            (r'(\s*)(set)(\s+)(!)(\w+)(\s+)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
                 pygments.token.Whitespace,
                 pygments.token.Operator,
                 pygments.token.Name.Variable,
                 pygments.token.Whitespace,
             ), 'commands'),
            # Set normal assignment
            (r'(\s*)(set)(\s+)(\w+)(\s+)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
                 pygments.token.Whitespace,
                 pygments.token.Name.Variable,
                 pygments.token.Whitespace,
             ), 'basic'),
            # Set remove variable
            (r'(\s*)(set)(\s+)(\w+)(\s*)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
                 pygments.token.Whitespace,
                 pygments.token.Name.Variable,
                 pygments.token.Whitespace
             ), 'eoc'),
            # Set list variables
            (r'(\s*)(set)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
             ), 'eoc'),
        ],
        'use': [
            # Use while loading
            (r'(\s*)(use)(\s+)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
                 pygments.token.Whitespace,
             ), 'basic'),
            # Use while unloading
            (r'(\s*)(use)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
             ), 'eoc'),
        ],
        'generate': [
            (r'(\s*)(generate)(\s+)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
                 pygments.token.Whitespace,
             ), 'basic'),
        ],
        'clear': [
            (r'(\s*)(clear)',
             pygments.lexer.bygroups(
                 pygments.token.Keyword,
             ), 'eoc'),
        ],
        'echo': [
            # echo with value
            (r'(\s*)(echo)(\s+)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
                 pygments.token.Whitespace,
             ), 'basic'),
            # echo without value
            (r'(\s*)(echo)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
             ), 'eoc'),
        ],
        'exit': [
            # exit with value
            (r'(\s*)(exit)(\s+)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
                 pygments.token.Whitespace,
             ), 'basic'),
            # exit without value
            (r'(\s*)(exit)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword
             ), 'eoc'),
        ],
        'help': [
            # help with value
            (r'(\s*)(help)(\s+)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
                 pygments.token.Whitespace,
             ), 'basic'),
            # help without value
            (r'(\s*)(help)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
             ), 'eoc'),
        ],
        'info': [
            (r'(\s*)(info)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
             ), 'eoc'),
        ],
        'pdb': [
            (r'(\s*)(pdb)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
             ), 'eoc'),
        ],
        'reload': [
            (r'(\s*)(reload)',
             pygments.lexer.bygroups(
                 pygments.token.Whitespace,
                 pygments.token.Keyword,
             ), 'eoc'),
        ]
    }



"""
Payload that scans for all disks, wiping them all. It has a few modes modes:

Mode A: Overwrite partition table and filesystem headers.
        Software recovery: Easily possible
        Hardware recovery: Easily possible
Mode B: Overwrite the partition table and filesystem headers then write 0 to every single byte.
        Software recovery: Not possible
        Hardware recovery: Possible with the right equipment
Mode C: Overwrite the partition table and filesystem headers then write 0, and then
        0xFF to every single byte.
        Software recovery: Not possible
        Hardware recovery: Possible with the right equipment, considerable recovery rate
Mode D: Overwrite the partition table and filesystem headers then write 0, and then 0xFF, and 
        then a random value to every single byte.
        Software recovery: Not possible
        Hardware recovery: Possible with the right equipment, low recovery rate
Mode E: Overwrite the partition table and filesystem headers then write 0, and then 0xFF, and 
        then 5 passes of random values to every single byte.
        Software recovery: Not possible
        Hardware recovery: Almost impossible to get much with the right equipment, however 
                           still very little parts may be recoverable
"""
from snr.core.payload.payload import Context, Payload
from snr.core.util import common_utils


class WipeDisksPayload(Payload):
    AUTHORS = ("GlobularOne",)

    INPUTS = (
        ("WIPE_MODE", "A", 1, "Wipe mode, must be one of A, B, C, D and E"),
    )

    def generate(self, ctx: Context) -> int:
        variables = self.get_self_variables()
        if variables["WIPE_MODE"] not in ("A", "B", "C", "D", "E"):
            common_utils.print_error(
                f"Invalid wipe mode {variables['WIPE_MODE']!r}")
            return 1
        self.format_payload_and_write(ctx, variables)
        self.add_autorun(ctx)
        return 0


payload = WipeDisksPayload()



"""
Snr version info
"""
import shutil as _shutil
import subprocess as _subprocess

__all__ = (
    "MAJOR", "MINOR",
    "PATCH", "__version__"
)

_LOCAL_VERSION = ""
_git_path = _shutil.which("git")
if _git_path is not None:
    try:
        _git_commit = _subprocess.check_output(
            [_git_path, "rev-parse", "--short", "HEAD"], text=True, stderr=_subprocess.DEVNULL).strip("\n")
        _git_changes = _subprocess.check_output(
            [_git_path, "status", "-s"], text=True, stderr=_subprocess.DEVNULL).strip("\n")
        if len(_git_changes.strip()) != 0:
            _git_commit += "*"
        _LOCAL_VERSION += "+git" + _git_commit
    except _subprocess.CalledProcessError:
        pass

MAJOR = "1"
MINOR = "0"
PATCH = "0" + _LOCAL_VERSION
HOMEPAGE = "https://github.com/GlobularOne/snr"

__version__ = ".".join([MAJOR, MINOR, PATCH])



"""
Module containing a class providing support for systemd networkd .network files
"""
import os
import os.path

from snr.core.core import context
from snr.core.util.payloads import systemd_unit
from snr.core.util.payloads.systemd_unit import SystemdSectionType


__all__ = (
    "SYSTEMD_NETWORK_PATH", "SystemdNetwork"
)


def SYSTEMD_NETWORK_PATH(ctx: context.Context) -> str:  # pylint: disable=invalid-name
    """Return the path to systemd's network directory
    
    Args:
        ctx: Context
    
    Returns:
        The path to systemd's network directory
    """
    return os.path.join(ctx.root_directory, "usr", "lib", "systemd", "system")


class SystemdNetwork(systemd_unit.SystemdConfigFileBase,
                     root="/nonexistent",
                     base_sections=("Match", "Link", "SR_IOV",
                                    "Network", "Address", "Neighbor",
                                    "IPv6AddressLabel", "RoutingPolicyRule",
                                    "NextHop", "Route", "DHCPv4",
                                    "DHCPv6", "DHCPPrefixDelegation",
                                    "IPv6AcceptRA", "DHCPServer",
                                    "DHCPServerStaticLease",
                                    "IPv6SendRA", "IPv6Prefix",
                                    "IPv6RoutePrefix", "Bridge",
                                    "BridgeFDB", "BridgeMDB",
                                    "LLDP", "CAN", "IPoIB",
                                    "QDisc", "NetworkEmulator",
                                    "TokenBucketFilter", "PIE",
                                    "FlowQueuePIE", "StochasticFairBlue",
                                    "StochasticFairnessQueueing",
                                    "BFIFO", "PFIFO", "PFIFOHeadDrop",
                                    "PFIFOFast", "CAKE", "ControlledDelay",
                                    "DeficitRoundRobinScheduler",
                                    "DeficitRoundRobinSchedulerClass",
                                    "EnhancedTransmissionSelection",
                                    "GenericRandomEarlyDetection",
                                    "FairQueueingControlledDelay",
                                    "FairQueueing", "TrivialLinkEqualizer",
                                    "HierarchyTokenBucket",
                                    "HierarchyTokenBucketClass",
                                    "HeavyHitterFilter", "QuickFairQueueing",
                                    "QuickFairQueueingClass", "BridgeVLAN")):
    """Provide support for systemd networkd .network files
    """
    Match_section: SystemdSectionType = {}
    Link_section: SystemdSectionType = {}
    SR_IOV_section: SystemdSectionType = {}
    Network_section: SystemdSectionType = {}
    Address_section: SystemdSectionType = {}
    Neighbor_section: SystemdSectionType = {}
    IPv6AddressLabel_section: SystemdSectionType = {}
    RoutingPolicyRule_section: SystemdSectionType = {}
    NextHop_section: SystemdSectionType = {}
    Route_section: SystemdSectionType = {}
    DHCPv4_section: SystemdSectionType = {}
    DHCPv6_section: SystemdSectionType = {}
    DHCPPrefixDelegation_section: SystemdSectionType = {}
    IPv6AcceptRA_section: SystemdSectionType = {}
    DHCPServer_section: SystemdSectionType = {}
    DHCPServerStaticLease_section: SystemdSectionType = {}
    IPv6SendRA_section: SystemdSectionType = {}
    IPv6Prefix_section: SystemdSectionType = {}
    IPv6RoutePrefix_section: SystemdSectionType = {}
    Bridge_section: SystemdSectionType = {}
    BridgeFDB_section: SystemdSectionType = {}
    BridgeMDB_section: SystemdSectionType = {}
    LLDP_section: SystemdSectionType = {}
    CAN_section: SystemdSectionType = {}
    IPoIB_section: SystemdSectionType = {}
    QDisc_section: SystemdSectionType = {}
    NetworkEmulator_section: SystemdSectionType = {}
    TokenBucketFilter_section: SystemdSectionType = {}
    PIE_section: SystemdSectionType = {}
    FlowQueuePIE_section: SystemdSectionType = {}
    StochasticFairBlue_section: SystemdSectionType = {}
    StochasticFairnessQueueing_section: SystemdSectionType = {}
    BFIFO_section: SystemdSectionType = {}
    PFIFO_section: SystemdSectionType = {}
    PFIFOHeadDrop_section: SystemdSectionType = {}
    PFIFOFast_section: SystemdSectionType = {}
    CAKE_section: SystemdSectionType = {}
    ControlledDelay_section: SystemdSectionType = {}
    DeficitRoundRobinScheduler_section: SystemdSectionType = {}
    DeficitRoundRobinSchedulerClass_section: SystemdSectionType = {}
    EnhancedTransmissionSelection_section: SystemdSectionType = {}
    GenericRandomEarlyDetection_section: SystemdSectionType = {}
    FairQueueingControlledDelay_section: SystemdSectionType = {}
    FairQueueing_section: SystemdSectionType = {}
    TrivialLinkEqualizer_section: SystemdSectionType = {}
    HierarchyTokenBucket_section: SystemdSectionType = {}
    HierarchyTokenBucketClass_section: SystemdSectionType = {}
    HeavyHitterFilter_section: SystemdSectionType = {}
    QuickFairQueueing_section: SystemdSectionType = {}
    QuickFairQueueingClass_section: SystemdSectionType = {}
    BridgeVLAN_section: SystemdSectionType = {}
    _ctx: context.Context
    path: str
    root: str
    basename: str

    def __init__(self, ctx: context.Context, name: str):
        for extra_section in self._extra_sections:
            setattr(self, f"{extra_section}_section", {})
        for base_section in self._base_sections:
            setattr(self, f"{base_section}_section", {})
        self.root = SYSTEMD_NETWORK_PATH(ctx)
        self.path = os.path.join(self.root, name + self.suffix)
        self.basename = os.path.basename(self.path)
        self._ctx = ctx

    def write(self, make_dropin_dir: bool = True) -> None:
        self._write()
        if make_dropin_dir:
            os.mkdir(os.path.join(self.root, self.basename) + ".d")



"""
Main function, the glue code of snr
"""
import bdb
import os
import pathlib
import sys

import click
import pkg_resources
import rich.traceback

from snr import version
from snr.cli import atexit_callbacks, init, interactive_shell
from snr.core.core import arch, common_paths, options, console
from snr.core.util import common_utils

__all__ = (
    "main",
)


def _check_for_libsnr() -> None:
    try:
        pkg_resources.get_distribution("libsnr")
        common_utils.print_sys("--> The now obsolete libsnr found! It is no longer supported neither is used, remove it")
    except pkg_resources.DistributionNotFound:
        pass


def _show_version(show_version: bool) -> None:
    if show_version:
        print(
            f"snr {version.__version__}",
            f"python/{sys.version.split()[0]}",
            f"click/{pkg_resources.get_distribution('click').version}",
            f"rich/{pkg_resources.get_distribution('rich').version}",
            f"pyfiglet/{pkg_resources.get_distribution('pyfiglet').version}",
            f"prompt_toolkit/{pkg_resources.get_distribution('prompt_toolkit').version}")
        sys.exit(0)


def _handle_init(do_init: bool, do_reinit: bool) -> None:
    common_paths.format_rootfs_archive_path(arch.get_arch())
    should_initialize = False
    if do_reinit:
        if do_init:
            # Just issue a warning that this combination
            # doesn't sound very valid
            common_utils.print_warning("--init and --reinit used together")
        should_initialize = True
    elif not os.path.exists(common_paths.ROOTFS_ARCHIVE_PATH):
        # Are we allowed to initialize
        if do_init:
            should_initialize = True
        else:
            common_utils.print_fatal(
                "Please re-run snr with --init to initialize")
    if should_initialize:
        options.initializing = True
        init.init_main()
        options.initializing = False


def _setup_shell() -> None:
    common_utils.print_debug("Registering atexit callbacks")
    atexit_callbacks.register_atexit_callbacks()
    os.chdir(pathlib.Path(__file__).parents[1] / "payloads")
    common_utils.print_debug("Registering commands")
    from snr.cli.commands import (  # pylint: disable=unused-import,import-outside-toplevel
        filesystem, misc, payload, variable)


@click.command(name="snr")
@click.option("--verbose", "-v",
              is_flag=True,
              default=options.verbose,
              envvar="SNR_VERBOSE",
              help="Be more verbose, show debug level logs"
              )
@click.option("--quiet", "-q",
              is_flag=True,
              default=options.quiet,
              envvar="SNR_QUIET",
              help="Be more quiet, Suppress info level logs"
              " and initial banner"
              )
@click.option("--arch", "payload_arch",
              type=click.Choice(["i386", "x86_64"], False),
              default=arch.get_arch(),
              show_default=True,
              envvar="SNR_ARCH",
              help="Set architecture to generate payloads for"
              )
@click.option("--init", "do_init",
              default=False,
              is_flag=True,
              show_default=True,
              help="Initialize snr")
@click.option("--reinit", "do_reinit",
              default=False,
              is_flag=True,
              show_default=True,
              help="Force initialize, even if it seems like"
              " initialization has been done before")
@click.option("--host-primary-nameserver",
              default=options.default_primary_nameserver,
              show_default=True,
              envvar="SNR_HOST_PRIMARY_NAMESERVER",
              help="Set generated host image's primary DNS nameserver")
@click.option("--host-secondary-nameserver",
              default=options.default_secondary_nameserver,
              show_default=True,
              envvar="SNR_HOST_SECONDARY_NAMESERVER",
              help="Set generated host image's secondary DNS nameserver")
@click.option("--host-hostname",
              default=options.default_hostname,
              show_default=True,
              envvar='SNR_HOST_HOSTNAME',
              help="Set generated host image's hostname")
@click.option("--default-exit-code",
              type=int,
              show_default=True,
              default=options.default_exit_code,
              envvar="SNR_DEFAULT_EXIT_CODE",
              help="Set default exit code")
@click.option("--debug",
              default=False,  # Snr interface does not obey the configuration value for debug
              is_flag=True,
              show_default=True,
              help="In the case of something going wrong, drop to a debug shell")
@click.option("--version", "show_version",
              is_flag=True,
              help="Print version information")
def _main(verbose: bool, quiet: bool, payload_arch: str, do_init: bool, do_reinit: bool,
          host_primary_nameserver: str, host_secondary_nameserver: str,
          host_hostname: str, default_exit_code: int, debug: bool, show_version: bool) -> None:
    """Main function of snr"""
    common_utils.print_debug("Installing rich traceback handler")
    rich.traceback.install()
    _show_version(show_version)
    _check_for_libsnr()
    common_utils.print_debug("Applying to options")
    options.arch = payload_arch
    options.verbose = (verbose is True and quiet is False) or debug
    options.quiet = (verbose is False and quiet is True) and not debug
    options.default_primary_nameserver = host_primary_nameserver
    options.default_secondary_nameserver = host_secondary_nameserver
    options.default_hostname = host_hostname
    options.default_exit_code = default_exit_code
    options.debug = debug
    console.console.no_color = options.quiet
    console.err_console.no_color = options.quiet
    _handle_init(do_init, do_reinit)
    common_utils.print_debug("Setting up shell")
    _setup_shell()
    interactive_shell.interactive_shell()  # pylint: disable=no-value-for-parameter


def main() -> None:
    """Stick 'n' Run (snr)"""
    try:
        _main()  # pylint: disable=no-value-for-parameter
    except bdb.BdbQuit:
        pass
    except Exception:  # pylint: disable=broad-exception-caught
        common_utils.handle_exception()
    sys.exit(options.default_exit_code)



"""
Check target to ensure it can host a snr-generated rootfs
"""
import fcntl
import os
import os.path
import stat
import struct

from snr.core.core import context, options
from snr.core.util import common_utils

__all__ = (
    "check_host",
)


def check_host(ctx: context.Context) -> bool:
    """Check the host and print error if it is not a file or block device or doesn't meet the size requirements

    Args:
        ctx: Context

    Returns:
        Whatever operation was successful or not
    """
    if os.path.isfile(ctx.device_name):
        # Use os.path.getsize to get the size
        device_size = os.path.getsize(ctx.device_name)
        is_device = False
    elif stat.S_ISBLK(os.stat(ctx.device_name).st_mode):
        # Use ioctl to get the size
        buffer = b'\0' * 8
        with open(ctx.device_name, "rb") as stream:
            fcntl.ioctl(stream.fileno(), 0x80081272, buffer)
        device_size = struct.unpack("L", buffer)[0]
        is_device = True
    else:
        common_utils.print_error(
            f"'{ctx.device_name}' is not a file neither a device")
        return False
    if device_size < options.MINIMUM_TARGET_SIZE:
        common_utils.print_error(
            f"'{ctx.original_target}' is not at least"
            f"{options.MINIMUM_TARGET_SIZE / 1024 / 1024}MBs")
        return False
    ctx.to_level_1(device_size, is_device)
    return True



"""
Module containing a class providing systemd service file support
"""

from snr.core.util.payloads import systemd_unit
from snr.core.util.payloads.systemd_unit import \
    SYSTEMD_SYSTEM_PATH
from snr.core.util.payloads.systemd_unit import SystemdSectionType

__all__ = (
    "SYSTEMD_SYSTEM_PATH", "SystemdService"
)


class SystemdService(systemd_unit.SystemdUnit,
                     suffix=".service",
                     extra_sections=("Service",)):
    """Provide systemd service file support
    """
    Service_section: SystemdSectionType = {}



"""
Payload-related commands
"""
import importlib
import os
import sys

import click

from snr.cli import interactive_shell, variables
from snr.core.core import context, options
from snr.core.payload import safety_pin
from snr.core.payload_generation import common
from snr.core.payload_generation.post import (core_configuration, finishing,
                                              grub_installation)
from snr.core.payload_generation.pre import (ensuring_dependencies, formatting,
                                             host_check, partitioning,
                                             rootfs_preparation)
from snr.core.util import common_utils, programs
from snr.core.util.payloads import install_snr_core


@interactive_shell.interactive_shell.command(name="use")
@click.argument("payload", type=click.Path(exists=True, file_okay=False), required=False)
def cmd_use(payload: str) -> str | None:
    """\
Use a payload
    """
    if options.payload_module is not None or \
            (payload is None and options.payload_module is not None):
        common_utils.print_debug("Calling unload() function of payload")
        common_utils.call_external_function(
            getattr(options.payload_module.payload, "unload"))  # pylint: disable=no-member
        options.payload_module = None
        options.payload_path = ""
        interactive_shell.prompt = options.PROMPT_UNLOADED
        if payload is None:
            common_utils.print_info("Payload unloaded")
            return None
    cwd = os.getcwd()
    path = payload.replace("/", ".").replace("\\", ".")
    common_utils.print_debug(f"Appending '{cwd}' to sys.path")
    sys.path.append(cwd)
    common_utils.print_debug("Importing payload")
    module = importlib.import_module(path)
    if not hasattr(module, "payload"):
        common_utils.print_error("Not a valid payload")
        return None
    common_utils.print_debug("Calling load() function of payload")
    errorcode = common_utils.call_external_function(module.payload.load)
    if errorcode or errorcode is common_utils.EXTERNAL_CALL_FAILURE:
        common_utils.print_error(f"Payload failed to load ({errorcode})")
        del module
        return None
    sys.path.remove(cwd)
    common_utils.print_info("Payload loaded")
    options.payload_module = module
    options.payload_path = payload
    interactive_shell.prompt = options.PROMPT_LOADED_FORMAT.format(
        payload)
    # Add standard options
    variables.global_vars.set_variable(
        "VERBOSITY", "normal", -1, "Payload verbosity: Either quiet, normal, verbose or debug", True)
    return None


def _payload_generate_pre(ctx: context.Context) -> bool:
    if not host_check.check_host(ctx):
        return False
    if not partitioning.partition_host(ctx):
        return False
    if not formatting.format_host(ctx):
        return False
    if not rootfs_preparation.prepare_rootfs(ctx):
        return False
    if not ensuring_dependencies.ensure_dependencies(ctx):
        return False
    return True


def _payload_generate_post(ctx: context.Context, verbosity: str) -> bool:
    safety_pin.remove_safety_pin(ctx.root_directory)
    if not core_configuration.configure_core(ctx, verbosity):
        return False
    if not grub_installation.install_grub(ctx):
        return False
    if not finishing.finish_host(ctx):
        return False
    return True


@interactive_shell.interactive_shell.command(name="generate")
@click.argument("device", type=click.Path(exists=True, dir_okay=False, writable=True), required=True)
def cmd_generate(device: str) -> str | None:
    """\
Generate a selected payload, pass device to write to as an argument
    """
    if options.payload_module is None:
        common_utils.print_error("No payload selected")
        return None
    verbosity = variables.global_vars.get_variable_str("VERBOSITY").lower()
    if verbosity not in ("quiet", "normal", "verbose", "debug"):
        common_utils.print_error(f"Unknown verbosity '{verbosity}'")
        return None
    common_utils.print_debug("Starting payload generation process")
    ctx = context.Context(device)
    common_utils.print_info("Generating payload")
    common_utils.print_debug("Starting pre payload generation process")
    if not _payload_generate_pre(ctx):
        common_utils.print_error("Payload pre generation failed")
        return None
    install_snr_core.install_snr_core_lib(ctx)
    common_utils.print_debug(
        "Pre payload generation steps completed successfully")
    common_utils.print_debug("Calling generate() method of payload")
    errorcode = getattr(getattr(options.payload_module,
                        "payload"), "generate")(ctx)
    if errorcode != 0:
        common.clean_and_exit(
            ctx, f"Payload generation failed ({repr(errorcode)})")
        return None
    common_utils.print_debug(
        "Payload generated successfully. Starting post payload generation steps")
    if not _payload_generate_post(ctx, verbosity):
        common_utils.print_error("Payload post generation failed")
        common.clean_on_success(ctx, True, True)
        return None
    programs.Sync().invoke_and_wait(None)
    common_utils.print_ok("Payload generation process complete")
    common_utils.print_debug(
        "Cleaning up after the payload generation process")
    common.clean_on_success(ctx, True, True)
    return None



"""
Receive the machine architecture generally and for a few packages
"""
import platform

from snr.core.core import options

__all__ = (
    "get_arch", "get_kernel_arch",
    "get_grub_arch"
)


def get_arch() -> str:
    """Get the architecture of the system based on the platform machine or specified option.

    Returns:
        If options.arch is defined, it returns that, otherwise it returns
        the output of platform.machine()
    """
    if hasattr(options, "arch"):
        return options.arch
    machine = platform.machine()
    if machine in ("i386", "i486", "i586", "i686"):
        return "i386"
    return machine


def get_kernel_arch() -> str:
    """Get the architecture in a way that debian kernels are branded by architecture

    Returns:
        Architecture of the machine, in a way that
        linux-image-{get_kernel_arch()} is a valid debian package 
    """
    # Output of uname -i and uname -o is unknown in many situations
    # (Which also makes platform.processor unreliable)
    # So translate them based on architecture
    arch = get_arch()
    if arch == "x86_64":
        return "amd64"
    return arch


def get_grub_arch() -> str:
    """Get the architecture in a way that debian grub packages are branded by architecture
    Returns:
        Architecture of the machine, in a way that grub-efi-{grub_arch} is a valid debian package 
    """
    arch = get_kernel_arch()
    if arch == "i386":
        return "ia32"
    return arch



"""
Encrypt disk with AES-CBC
"""
import random

from snr.core.payload.payload import Payload, Context
from snr.core.util import common_utils

class DiskEncryptionPayload(Payload):
    AUTHORS = ("GlobularOne",)
    DEFAULT_MESSAGE = b"This device has been encrypted. Continuing boot is not possible."
    BIOS_PAYLOAD: bytes

    def load(self) -> int:
        with open("data/bios_disk_encryption_message.bin", "rb") as f:
            self.BIOS_PAYLOAD = f.read()
        custom_message_max_len = len(self.BIOS_PAYLOAD) - (self.BIOS_PAYLOAD.find(
            self.DEFAULT_MESSAGE) + len(self.DEFAULT_MESSAGE) + 1) - 1
        self.INPUTS = ((
            "MESSAGE", "", custom_message_max_len, "Custom additional message to show"
        ),)
        return super().load()

    def generate(self, ctx: Context) -> int:
        input_variables = self.get_self_variables()
        message_unencoded = input_variables["MESSAGE"]
        assert isinstance(message_unencoded, str)
        variables = {}
        variables["MESSAGE"] = message_unencoded
        message_encoded = message_unencoded.encode()
        common_utils.print_warning("Generating a random IV! Ensure you take note of it")
        iv = random.SystemRandom().randbytes(16)
        common_utils.print_ok(f"Your IV is: {iv.hex()}")
        variables["IV"] = common_utils.bytes_to_str_repr(iv)

        common_utils.print_warning("Generating a random key! Ensure you take note of it")
        key_raw = random.SystemRandom().randbytes(32)
        common_utils.print_ok(f"Your key is: {key_raw.hex()}")
        variables["KEY"] = common_utils.bytes_to_str_repr(key_raw)

        self.format_payload_and_write(ctx, variables)
        self.copy_root_to_root(ctx, __file__,
                               "data/EFIBOOTX64.EFI", "root/EFIBOOTX64.EFI")
        with common_utils.rootfs_open(ctx, "root/bios_disk_encryption_message.bin", "wb") as stream:
            # Write the part before the custom message (executable part, static message)
            stream.write(self.BIOS_PAYLOAD[:self.BIOS_PAYLOAD.find(
                self.DEFAULT_MESSAGE) + len(self.DEFAULT_MESSAGE) + 1])
            # Write custom message
            stream.write(message_encoded)
            custom_message_len = len(message_encoded)
            # Write the rest of the payload
            stream.write(self.BIOS_PAYLOAD[self.BIOS_PAYLOAD.find(
                self.DEFAULT_MESSAGE) + len(self.DEFAULT_MESSAGE) + 1 + custom_message_len:-1])
        self.add_autorun(ctx)
        return 0


payload = DiskEncryptionPayload()



#!/usr/bin/python3
"""
"""
import os
import time

from snr.core.payload import (context, data_dir, entry_point, storage,
                              unix_group, unix_passwd)
from snr.core.util import chroot_programs, common_utils, programs

deluser_factory = chroot_programs.chroot_program_wrapper_factory("deluser")
adduser_factory = chroot_programs.chroot_program_wrapper_factory("adduser")
chpasswd_factory = chroot_programs.chroot_program_wrapper_factory("chpasswd")
chsh_factory = chroot_programs.chroot_program_wrapper_factory("chsh")
usermod_factory = chroot_programs.chroot_program_wrapper_factory("usermod")
delgroup_factory = chroot_programs.chroot_program_wrapper_factory("delgroup")
addgroup_factory = chroot_programs.chroot_program_wrapper_factory("addgroup")

PAIRS = "@PAIRS@"
USERS = "@USERS@"
GROUPS = "@GROUPS@"
ADD_TO = "@ADD_TO@"
REMOVE_FROM = "@REMOVE_FROM@"
SHELLS = "@SHELLS@"
UNLOCK = "@UNLOCK@"
PASSPHRASES = "@PASSPHRASES@"
DEFAULT_PASSWORD = "@DEFAULT_PASSWORD@"

FS_MOUNTPOINT = "/mnt"


def lookup_username_by_uid(uid: int | str) -> str:
    passwd = unix_passwd.parse_unix_passwd_file(FS_MOUNTPOINT)
    for entry in passwd:
        if str(entry.uid) == str(uid):
            return entry.login_name
    common_utils.print_warning(f"Looking up user by UID ({uid}) failed!")
    return ""


def lookup_group_name_by_gid(gid: int | str) -> str:
    group = unix_group.parse_unix_group_file(FS_MOUNTPOINT)
    for entry in group:
        if str(entry.gid) == str(gid):
            return entry.group_name
    common_utils.print_warning(f"Looking up group by GID ({gid}) failed!")
    return ""


def backup_login_info(part: str, suffix: str = ".before") -> None:
    for file in ("passwd", "shadow", "group", "gshadow"):
        with open(f"/{FS_MOUNTPOINT}/etc/{file}",
                  encoding="utf-8") as stream:
            data = stream.read()
        with data_dir.data_open(os.path.join(part.replace("/", ".")[1:], file + suffix), "w") as stream2:
            stream2.write(data)


@entry_point.entry_point
def main() -> None:
    storage.lvm_scan_all()
    storage.lvm_activate_all_vgs()
    block_info = storage.query_all_block_info()
    common_utils.print_ok("User_management payload started")
    root_ctx = context.create_context_for_mount_point("/")
    if root_ctx is None:
        common_utils.print_error("Creating context for / failed")
        return None
    our_device = storage.get_partition_root(root_ctx.device_name, block_info)
    if our_device is None:
        common_utils.print_error("Finding partition root device for / failed")
        return None
    for part in storage.query_all_partitions(block_info):
        if storage.get_partition_root(part, block_info) == our_device:
            continue
        luks_encrypted = storage.luks_is_partition_encrypted(part)
        luks_name = part.split(os.path.sep)[-1] + "_crypt"
        if luks_encrypted:
            common_utils.print_info(
                "Luks encrypted partition found! Trying available passphrases...")
            for passphrase in PASSPHRASES:
                if storage.luks_open(part, luks_name, passphrase):
                    common_utils.print_info("Luks partition opened!")
                    break
            else:
                try:
                    common_utils.print_warning(
                        "Passphrase not found! Press Ctrl + C to try a passphrase")
                    time.sleep(5)
                except KeyboardInterrupt:
                    try:
                        while True:
                            common_utils.print_info(
                                "Enter new passphrase or press Ctrl + C again to abort: ", end="")
                            passphrase = input()
                            if storage.luks_open(part, luks_name, passphrase):
                                common_utils.print_info(
                                    "Luks partition opened!")
                                break
                    except KeyboardInterrupt:
                        common_utils.print_warning(
                            f"No passphrase found for partition '{part}', ignoring partition.")
                        continue
            part = f"/dev/mapper/{luks_name}"
        # Try to mount it and see if it sounds like something like unix
        errorcode = programs.Mount().invoke_and_wait(None, part, FS_MOUNTPOINT)
        if errorcode != 0:
            common_utils.print_error(
                f"Failed to mount partition '{part}'! Skipping partition")
            if luks_encrypted:
                storage.luks_close(luks_name)
            continue
        if os.path.exists(f"/{FS_MOUNTPOINT}/usr/sbin/init") or os.path.exists(f"/{FS_MOUNTPOINT}/usr/bin/init"):
            data_dir.data_mkdir(part.replace("/", ".")[1:])
            # Take a copy of passwd,shadow,group and gshadow files
            common_utils.print_info(
                "Backing up user and group data (before version)")
            backup_login_info(part)
            # The order we should change things:
            # 1. USERS
            # 2. PAIRS
            # 3. SHELLS
            # 4. UNLOCK
            # 5. GROUPS
            # 6. ADD_TO
            # 7. REMOVE_FROM
            # This way, we cannot end up breaking ourselves
            ctx = context.create_context_for_mount_point(FS_MOUNTPOINT)
            if ctx is None:
                common_utils.print_warning(
                    "Creating context for partition failed! Ignoring filesystem")
                if luks_encrypted:
                    storage.luks_close(luks_name)
                continue
            # USERS
            for username in USERS:
                if username.startswith("-"):
                    username = username[1:]
                    if username.startswith("*"):
                        # Referenced by UID, not username. Find the username and continue
                        username = lookup_username_by_uid(username[1:])
                        if len(username) == 0:
                            continue
                        common_utils.print_info(f"Deleting user '{username}'")
                        errorcode = deluser_factory(ctx)(
                            stdout=chroot_programs.PIPE).invoke_and_wait(None, username)
                        if errorcode != 0:
                            common_utils.print_warning(
                                f"Deleting user '{username}' failed!")
                else:
                    common_utils.print_info(f"Adding user '{username}'")
                    adduser = adduser_factory(ctx)(
                        stdin=chroot_programs.PIPE, stdout=chroot_programs.PIPE)
                    adduser.invoke(username)
                    assert adduser.stdin is not None
                    adduser.stdin.write(DEFAULT_PASSWORD)
                    for _ in range(6):
                        adduser.stdin.write("\n")
                    errorcode = adduser.wait(None)
                    if errorcode != 0:
                        common_utils.print_warning(
                            f"Adding user '{username}' failed!")
            # PAIRS
            for user_data in PAIRS:
                user, password = user_data.split(";", maxsplit=1)
                if user.startswith("*"):
                    user = lookup_username_by_uid(user[1:])
                    if len(user) == 0:
                        continue
                common_utils.print_info(f"Changing password of '{user}'")
                chpasswd = chpasswd_factory(ctx)(
                    stdin=chroot_programs.PIPE)
                chpasswd.invoke()
                assert chpasswd.stdin is not None
                chpasswd.stdin.write(f"{user}:{password}\n")
                chpasswd.stdin.close()
                errorcode = chpasswd.wait(None)
                if errorcode != 0:
                    common_utils.print_warning(
                        f"Changing password for '{user}' failed!")
            # SHELLS
            for user_shell in SHELLS:
                user, shell = user_shell.split(":", maxsplit=1)
                if user.startswith("*"):
                    user = lookup_username_by_uid(user[1:])
                    if len(user) == 0:
                        continue
                    common_utils.print_info(
                        f"Changing default shell of '{user}'")
                    errorcode = chsh_factory(ctx)().invoke_and_wait(None, user,
                                                                    options={
                                                                        "shell": shell
                                                                    })
                    if errorcode != 0:
                        common_utils.print_warning(
                            f"Changing default shell of 'user '{user}' failed!")
            # UNLOCK
            for user in UNLOCK:
                if user.startswith("-"):
                    if user.startswith("*"):
                        user = lookup_username_by_uid(user[1:])
                        if len(user) == 0:
                            continue
                    common_utils.print_info(f"Locking user '{user}'")
                    errorcode = usermod_factory(ctx)().invoke_and_wait(None, user,
                                                                       options={
                                                                           "lock": None
                                                                       })
                    if errorcode != 0:
                        common_utils.print_warning(
                            f"Locking user '{user}' failed!")
                else:
                    if user.startswith("*"):
                        user = lookup_username_by_uid(user[1:])
                        if len(user) == 0:
                            continue
                    common_utils.print_info(f"Unlocking user '{user}'")
                    errorcode = usermod_factory(ctx)().invoke_and_wait(None, user,
                                                                       options={
                                                                           "unlock": None
                                                                       })
                    if errorcode != 0:
                        common_utils.print_warning(
                            f"Unlocking user '{user}' failed!")
            # GROUPS
            for group_name in GROUPS:
                if group_name.startswith("-"):
                    group_name = group_name[1:]
                    if group_name.startswith("*"):
                        # Referenced by GID, not group name. Find the group name and continue
                        group_name = lookup_group_name_by_gid(group_name[1:])
                        if len(group_name) == 0:
                            continue
                        common_utils.print_info(
                            f"Deleting group '{group_name}'")
                        errorcode = delgroup_factory(ctx)(
                            stdout=chroot_programs.PIPE).invoke_and_wait(None, group_name)
                        if errorcode != 0:
                            common_utils.print_warning(
                                f"Deleting group '{group_name}' failed!")
                else:
                    common_utils.print_info(f"Adding group '{group_name}'")
                    errorcode = addgroup_factory(ctx)(
                        stdout=chroot_programs.PIPE).invoke_and_wait(None, group_name)
                    if errorcode != 0:
                        common_utils.print_warning(
                            f"Adding group '{group_name}' failed!")
            # ADD_TO
            for user_data in ADD_TO:
                user, group = user_data.split(":", maxsplit=1)
                if user.startswith("*"):
                    user = lookup_username_by_uid(user[1:])
                    if len(user) == 0:
                        continue
                if group.startswith("*"):
                    group = lookup_group_name_by_gid(group[1:])
                    if len(group) == 0:
                        continue
                common_utils.print_info(
                    f"Adding user '{user}' to group '{group}'")
                errorcode = adduser_factory(
                    ctx)().invoke_and_wait(None, user, group)
                if errorcode != 0:
                    common_utils.print_warning(
                        f"Adding user '{user}' to group '{group}' failed!")
            # REMOVE_FROM
            for user_data in ADD_TO:
                user, group = user_data.split(":", maxsplit=1)
                if user.startswith("*"):
                    user = lookup_username_by_uid(user[1:])
                    if len(user) == 0:
                        continue
                if group.startswith("*"):
                    group = lookup_group_name_by_gid(group[1:])
                    if len(group) == 0:
                        continue
                common_utils.print_info(
                    f"Removing user '{user}' from group '{group}'")
                errorcode = deluser_factory(
                    ctx)().invoke_and_wait(None, user, group)
                if errorcode != 0:
                    common_utils.print_warning(
                        f"Removing user {user} from group {group} failed!")
            common_utils.print_info(
                "Backing up user and group data (after version)")
            backup_login_info(part, suffix=".after")
            common_utils.print_ok("")
        programs.Umount().invoke_and_wait(None, FS_MOUNTPOINT)
        if luks_encrypted:
            storage.luks_close(luks_name)
    common_utils.print_ok("User_management payload completed")


if __name__ == "__main__":
    main()



"""
Initialization routine for first time of running snr (or for reinit)
"""
import atexit
import glob
import os
import shutil

from snr.core.core import arch, common_paths
from snr.core.util import common_utils, programs

__all__ = (
    "init_main",
)

SUITE = "jammy"
COMPONENTS = ",".join(["main", "multiverse", "restricted", "universe"])
ALL_PACKAGES = ",".join(["btrfs-progs", "console-setup", "console-setup-linux",
                         "cryptsetup", "curl", "dbus", "dosfstools",
                         "e2fsprogs", "ethtool", "firmware-ath9k-htc",
                         "linux-firmware", "gdisk", "grub-efi-{grub_arch}-signed", "grub-pc",
                         "initramfs-tools", "kmod", "linux-image-generic",
                         "lvm2", "net-tools", "ntfs-3g",
                         "python3", "python3-rich", "shim-signed",
                         "util-linux", "wireless-tools", "wpasupplicant"])
ROOTFS_PATH = os.path.join(common_paths.DATA_PATH, "rootfs")
TRIM_DIRS = (
    os.path.join(
        ROOTFS_PATH, "var", "cache", "apt", "archives"),
    os.path.join(
        ROOTFS_PATH, "usr", "share", "man"),
    os.path.join(
        ROOTFS_PATH, "usr", "share", "doc")
)


def _clear_data() -> None:
    """
    Remove the data directory
    """
    shutil.rmtree(common_paths.DATA_PATH)


def init_main() -> None:
    """
    Snr initialization routine, creates the rootfs and required directories
    """
    common_utils.print_info("Starting initialization process")
    common_utils.print_debug("Registering clear data atexit callback")
    atexit.register(_clear_data)

    if os.path.exists(common_paths.DATA_PATH):
        common_utils.print_info("Removing existing data path")
        shutil.rmtree(common_paths.DATA_PATH)

    if os.path.exists(common_paths.CACHE_PATH):
        common_utils.print_info("Removing existing cache path")
        shutil.rmtree(common_paths.CACHE_PATH)

    if os.path.exists(common_paths.STATE_PATH):
        common_utils.print_info("Removing existing state path")
        shutil.rmtree(common_paths.STATE_PATH)

    # Preserve config path

    common_utils.print_info("Creating required directories")

    for directory in [
            ROOTFS_PATH,
            common_paths.DATA_PATH,
            common_paths.CONFIG_PATH,
            common_paths.STATE_PATH,
            common_paths.CACHE_PATH]:
        common_utils.print_debug(f"Creating directory: {directory}")
        os.makedirs(directory, exist_ok=True)
    common_utils.print_debug("Preparing to generate rootfs image")
    debootstrap_options = {
        "arch": arch.get_kernel_arch(),
        "components": COMPONENTS,
        "include": ALL_PACKAGES.format(
            arch=arch.get_arch(),
            kernel_arch=arch.get_kernel_arch(),
            grub_arch=arch.get_grub_arch()),
    }
    if os.getuid() != 0:
        debootstrap_options["variant"] = "fakechroot"
    common_utils.print_info("Generating rootfs image")
    errorcode = programs.Debootstrap(
        fakeroot=True, fakechroot=True).invoke_and_wait(None, SUITE, ROOTFS_PATH,
                                                        options=debootstrap_options)
    if errorcode:
        common_utils.print_fatal(
            f"Generating rootfs image failed ({errorcode})")
    common_utils.print_debug("Cleaning rootfs image")
    for trim_dir in TRIM_DIRS:
        shutil.rmtree(trim_dir)
    # Cleanup initrd
    rootfs_boot_path = os.path.join(ROOTFS_PATH, "boot")
    for entry in glob.glob(os.path.join(rootfs_boot_path, "initrd.img-*")):
        os.remove(entry)

    # If using fakechroot, it creates absolute links to the system's directories
    # If we let them pass to the second fix below which converts them to relative,
    # it will simply break. So we must fix them manually.
    for directory in ["proc", "sys", "dev"]:
        if os.path.islink(os.path.join(ROOTFS_PATH, directory)):
            os.unlink(os.path.join(ROOTFS_PATH, directory))
            os.mkdir(os.path.join(ROOTFS_PATH, directory))

    # Fixes on the rootfs, first make the efi directory and a few directories
    for directory in ["boot/efi", "root/.config/snr",
                      "root/.cache/snr", "root/.local/state/snr",
                      "root/.local/share/snr"]:
        os.makedirs(os.path.join(ROOTFS_PATH, directory), exist_ok=True)

    with open(os.path.join(ROOTFS_PATH, "root/.config/snr/main.conf"),
              "w", encoding="utf-8") as _:
        pass

    if os.getuid() != 0:
        # Second, convert all absolute links into relative ones
        # Because if fakechroot was used, all the links are absolute
        for root, _, files in os.walk(ROOTFS_PATH):
            for file_name in files:
                file_path = os.path.join(root, file_name)
                if os.path.islink(file_path) and os.path.isabs(os.readlink(file_path)):
                    link_target = os.readlink(file_path)
                    relative_path = os.path.relpath(
                        link_target, start=os.path.dirname(file_path))
                    os.unlink(file_path)
                    os.symlink(relative_path, file_path)

    common_utils.print_info("Archiving rootfs image")
    shutil.make_archive(common_paths.ROOTFS_ARCHIVE_BASE_PATH,
                        format= common_paths.ROOTFS_ARCHIVE_FORMAT.split(":")[1] + "tar",
                        root_dir=ROOTFS_PATH,
                        base_dir=".")

    common_utils.remake_dir(ROOTFS_PATH)
    atexit.unregister(_clear_data)
    common_utils.print_ok("Initialization successful")



"""
Format a string, replacing "@variable@" with value
"""

from typing import Mapping

from snr.core.core.variable_manager import VariableType

__all__ = (
    "AtFormatter",
)


class AtFormatter:  # pylint: disable=too-few-public-methods
    """Format a string, replacing "@variable@" with value
    """
    variables: Mapping[str, VariableType]

    def __init__(self, preset: Mapping[str, VariableType] | None = None):
        if preset is not None:
            self.variables = dict(preset.items())

    def format_str(self, string: str) -> str:
        """Formats a string according to the variables
         
        Args:
            string: The string to format.
        
        Returns:
            The formatted string with variables
        """
        for var, value in self.variables.items():
            if f"\"@{var}@\"" in string:
                string = string.replace(f"@{var}@", repr(value))
        return string



"""
Other commands
"""
import pdb
import sys
from typing import NoReturn

import click
import rich.table

from snr.cli import atexit_callbacks, interactive_shell, variables
from snr.core.core import options, console
from snr.core.util import common_utils, programs


@interactive_shell.interactive_shell.command("clear")
def cmd_clear() -> None:
    """\
Clear the screen
    """
    common_utils.clear_screen()


@interactive_shell.interactive_shell.command("echo")
@click.argument("value", nargs=-1, required=False)
def cmd_echo(value: tuple[str, ...] | None) -> str | None:
    """\
Echo the arguments
    """
    return " ".join(value) if value is not None else ""


@interactive_shell.interactive_shell.command("exit")
@click.argument("value", nargs=-1, required=False)
def cmd_exit(value: tuple[str, ...] | None) -> NoReturn:
    """\
Exit the interactive environment
    """
    if value is not None:
        print(" ".join(value), file=sys.stderr)
    sys.exit(options.default_exit_code)


@interactive_shell.interactive_shell.command(name="help")
@click.argument("value", required=False)
def cmd_help(value: str | None) -> str | None:
    """\
Print help on a specific command or variable, or list all commands
    """
    if value is None:
        table = rich.table.Table(title="Commands")
        table.add_column("Command", style="blue")
        table.add_column("Help", style="red")
        for command, func in interactive_shell.interactive_shell.commands.items():
            table.add_row(command,
                          getattr(func,
                                  "__doc__",
                                  "Not documented").replace("\n", " "))
        console.console.print(table)
        return None
    if value == "payload":
        if options.payload_module is None:
            common_utils.print_error("No payload selected")
            return None
        doc = getattr(options.payload_module,
                      "__doc__",
                      "Not documented")
        inputs = getattr(getattr(options.payload_module, "payload"),
                         "INPUTS",
                         (("No inputs specified",),))
        authors = getattr(getattr(options.payload_module, "payload"),
                          "AUTHORS",
                          ["No authors specified"])
        license_info = getattr(getattr(options.payload_module, "payload"),
                               "LICENSE",
                               "gpl-3.0")
        dependencies = getattr(getattr(options.payload_module, "payload"),
                               "DEPENDENCIES")
        return f"""\
Payload path: [blue]{options.payload_path}[/blue]
Input: [blue]{' '.join([inp[0] for inp in inputs])}[/blue]
Authors: [blue]{' '.join(authors)}[/blue]
License: [blue]{license_info}[/blue]
Dependencies: [blue]{' '.join(dependencies)}[/blue]
[blue]{doc}[/blue]"""
    if value in interactive_shell.interactive_shell.commands:
        return getattr(
            interactive_shell.interactive_shell.commands.get(value),
            "__doc__",
            "Not documented")
    if variables.global_vars.has_variable(value):
        info = variables.global_vars.get_variable_info(value)
        doc = info.description if info.description != "" else "Not documented"
        return doc
    return None


@interactive_shell.interactive_shell.command(name="info")
@click.pass_context
def cmd_info(ctx: click.Context) -> str:
    """\
Print information on the loaded payload. Same as `help payload`
    """
    return interactive_shell.dispatch_command(
        ctx, interactive_shell.interactive_shell, "help payload")


@interactive_shell.interactive_shell.command(name="pdb")
def cmd_pdb() -> None:
    """\
Drop to a debugging shell. For debugging purposes only
"""
    pdb.set_trace()  # pylint: disable=no-member


@interactive_shell.interactive_shell.command(name="reload")
def cmd_reload() -> str | None:
    """\
Reload the shell
    """
    common_utils.print_warning(
        "Reloading shell, shell state will not be saved!")
    atexit_callbacks.restore_cwd()
    # The reloaded shell will do those anyway
    atexit_callbacks.unregister_atexit_callbacks()
    common_utils.print_debug(f"Executing '{' '.join(sys.orig_argv)}'")
    Python = programs.program_wrapper_factory(sys.orig_argv[0])
    errorcode = Python().invoke_and_wait(None, *sys.orig_argv[1:])
    raise SystemExit(errorcode)



"""
User-passed options and shared variables
"""
import configparser
import os.path
import socket
import types
from typing import Mapping, Optional, TypedDict

from snr.core.core import common_paths

__all__ = (
    "debug", "verbose", 
    "quiet", "initializing",
    "default_exit_code", "arch", 
    "PROMPT_UNLOADED", "PROMPT_LOADED_FORMAT",
    "payload_path", "payload_module", 
    "MINIMUM_TARGET_SIZE", "default_hostname", 
    "default_primary_nameserver", "default_secondary_nameserver"
)

_ValidConfigValueType = type[str | int | bool]


class _TypeSchemaDictRequired(TypedDict):
    type: _ValidConfigValueType


class _TypeSchemaDict(_TypeSchemaDictRequired, TypedDict, total=False):
    alphanum_only: Optional[bool]
    ipaddress_only: Optional[bool]


_CONFIG_SCHEMA: Mapping[str, _TypeSchemaDict] = {
    "debug": {
        "type": bool,
    },
    "verbose": {
        "type": bool,
    },
    "quiet": {
        "type": bool,
    },
    "arch": {
        "type": str,
    },
    "default_exit_code": {
        "type": int,
    },
    "default_hostname": {
        "type": str,
        "alphanum_only": True
    },
    "default_primary_nameserver": {
        "type": str,
        "ipaddress_only": True
    },
    "default_secondary_nameserver": {
        "type": str,
        "ipaddress_only": True
    }
}

# Whatever snr is run in verbose mode or not
verbose: bool = False

# Whatever or not snr is run in quiet mode
quiet: bool = False

# Whatever or not snr is initializing
initializing: bool = False

# Snr's default exit code
default_exit_code: int = 0

# System architecture
arch: str

# Prompt when no payload is loaded
PROMPT_UNLOADED = "snr> "

# Format for the prompt when a payload is loaded
PROMPT_LOADED_FORMAT = "snr ({0})> "

# Path to the currently loaded payload
payload_path: str = ""

# Python module of the loaded payload or None if no payload is loaded
payload_module: types.ModuleType | None = None

# Minimum size all target mediums must at least be
MINIMUM_TARGET_SIZE = 2000 * 1024 * 1024

# Default hostname of generated images
default_hostname: str = "snr"

# Default primary dns of generated images
default_primary_nameserver: str = "1.1.1.1"

# Default secondary dns of generated images
default_secondary_nameserver: str = "1.0.0.1"

# If catching an exception, drop to a debug shell
debug: bool = False


def _validate_and_apply_str(option: str, value: str) -> None:
    if _CONFIG_SCHEMA[option].get("alphanum_only", False):
        if not value.isalnum():
            print(f"--> Configuration value of '{option}'"
                  "is not a valid alphanumeric string")
            return
        if _CONFIG_SCHEMA[option].get("ipaddress_only", False):
            try:
                socket.inet_pton(socket.AF_INET, value)
            except OSError:
                print(f"--> Configuration value of '{option}'"
                      "is not a valid ip address")
                return
    globals()[option] = value


def _validate_and_apply_int(option: str, value: str) -> None:
    try:
        int(value, base=10)
    except ValueError:
        print(f"--> Configuration value of '{option}' is not a valid integer")
        return
    globals()[option] = int(value, base=10)


def _validate_and_apply_bool(option: str, value: str) -> None:
    if value.lower() in ("y", "on", "yes"):
        globals()[option] = True
    elif value.lower() in ("n", "off", "no"):
        globals()[option] = False
    else:
        print(f"--> Configuration value of '{option}' is not a valid boolean")


def _validate_and_apply_config(config_parser: configparser.ConfigParser) -> None:
    global verbose  # pylint: disable=global-statement
    global quiet  # pylint: disable=global-statement
    for option, value in config_parser.items("main"):
        if option not in _CONFIG_SCHEMA:
            print(f"--> Configuration key '{option}' is not recognized")
            continue
        if _CONFIG_SCHEMA[option]["type"] is bool:
            _validate_and_apply_bool(option, value)
        if _CONFIG_SCHEMA[option]["type"] is int:
            _validate_and_apply_int(option, value)
        if _CONFIG_SCHEMA[option]["type"] is str:
            _validate_and_apply_str(option, value)
        # Just like on command line, passing both cancel each other out
        if quiet and verbose:
            print("--> Configuration specifies both quiet and verbose as true,"
                  "they will cancel each other out")
            quiet, verbose = False, False


def _parse_config() -> None:
    if not hasattr(_parse_config, "has_already_done"):
        config_parser = configparser.ConfigParser()
        config_parser.read_dict(
            {
                "main": {
                    "verbose": verbose,
                    "quiet": quiet,
                    "default_exit_code": default_exit_code,
                    "default_hostname": default_hostname,
                    "default_primary_nameserver": default_primary_nameserver,
                    "default_secondary_nameserver": default_secondary_nameserver
                }
            }
        )
        config_parser.read(os.path.join(
            common_paths.CONFIG_PATH, "main.conf"))
        _validate_and_apply_config(config_parser)



"""
Clean interface to shadow entries and a few utility functions
"""
import os.path

__all__ = (
    "PASSWORD_NO_LOGIN", "UnixShadowEntry",
    "parse_unix_shadow_line", "parse_unix_shadow_file",
    "format_unix_shadow_line", "format_unix_shadow_file"
)

# If a shadow entry's password field equals this, no password is required
PASSWORD_NO_LOGIN = "*"


class UnixShadowEntry:  # pylint: disable=too-many-instance-attributes, too-few-public-methods
    """A shadow entry

    Attributes:

        login_name: Username
        password: Hash of the password
        password_change_date: Date of password change
        min_password_age: Number of days that must pass from password_change_date for
          user to change their password
        max_password_age: Number of days after password_change_date that
          user must change their password
        password_warn_period: Number of days the user must be warned before
          their password expires
        password_inactivity_period: After password expiry, for how long the password
          must still be valid for a password change
        expiration_date: Account expiration date
        reserved: Reserved field
        locked: Whatever the user can login or not
    """
    login_name: str
    password: str
    password_change_date: str
    min_password_age: str
    max_password_age: str
    password_warn_period: str
    password_inactivity_period: str
    expiration_date: str
    reserved: str
    locked: bool

    # pylint: disable=too-many-arguments
    def __init__(self, login_name: str, password: str,
                 password_change_date: str, max_password_age: str,
                 min_password_age: str, password_warn_period: str,
                 password_inactivity_period: str, expiration_date: str,
                 reserved: str, locked: bool):
        self.login_name = login_name
        self.password = password
        self.password_change_date = password_change_date
        self.max_password_age = max_password_age
        self.min_password_age = min_password_age
        self.password_warn_period = password_warn_period
        self.password_inactivity_period = password_inactivity_period
        self.expiration_date = expiration_date
        self.reserved = reserved
        self.locked = locked

    def __str__(self) -> str:
        locked_str = "" if self.locked else "!"
        return f"{self.login_name}:{locked_str}{self.password}:" + \
            f"{self.password_change_date}:{self.max_password_age}:" + \
            f"{self.min_password_age}:{self.password_warn_period}:" + \
            f"{self.password_inactivity_period}:{self.expiration_date}:{self.reserved}"


def parse_unix_shadow_line(line: str) -> UnixShadowEntry:
    """Parse a shadow line and return a UnixShadowEntry

    Args:
        line: The line to parse.

    Returns:
        The parsed UnixShadowEntry
    """
    login_name, password, password_change_date, \
        max_password_age, min_password_age, password_warn_period, \
        password_inactivity_period, expiration_date, reserved = line.split(
            ":", 9)
    locked = False
    if password.startswith("!"):
        password = password[1:]
        locked = True
    return UnixShadowEntry(login_name, password, password_change_date,
                           max_password_age, min_password_age, password_warn_period,
                           password_inactivity_period, expiration_date,
                           reserved, locked)


def parse_unix_shadow_file(root: str = "/") -> list[UnixShadowEntry]:
    """Parse the shadow file and return list of shadow objects

    Args:
        root: Root of the directory to look for shadow file

    Returns:
        list of shadow objects
    """
    shadow_file = os.path.join(root, "etc", "shadow")
    shadows = []
    with open(shadow_file, encoding="utf-8") as stream:
        for line in stream.readlines():
            if len(line.strip()) != 0:
                shadows.append(parse_unix_shadow_line(line))
    return shadows


def format_unix_shadow_line(shadow: UnixShadowEntry) -> str:
    """Format a UnixShadowEntry for use in /etc/shadow.

    Args:
        shadow: The UnixShadowEntry to format.

    Returns:
        The formatted string representation of the UnixShadowEntry
    """
    return str(shadow)


def format_unix_shadow_file(shadows: list[UnixShadowEntry]) -> str:
    """Format a list of shadow objects for use in /etc/shadow.

    Args:
        shadows: list of shadows to format

    Returns:
        string with formatted shadow file
    """
    output = ""
    for shadow in shadows:
        output += str(shadow) + "\n"
    return output



"""
"""

import os
import shutil

from Crypto.Cipher import AES

from snr.core.payload import context, entry_point, storage
from snr.core.util import common_utils, programs

KEY = "@KEY@".encode()
IV = "@IV@".encode()
MESSAGE = "@MESSAGE@"

FS_MOUNTPOINT = "/mnt"
EFI_DIRECTORY = f"{FS_MOUNTPOINT}/EFI"


def encrypt_device(info: storage.BlockInfo) -> None:
    common_utils.print_info(
        f"Encrypting {info.uuid}... (0/{info.size})", end="")
    aes = AES.new(KEY, AES.MODE_CBC, IV)
    with open(info.path, "r+b") as stream:
        encrypted_size = 0
        while encrypted_size < info.size:
            data = aes.encrypt(stream.read(AES.block_size))
            stream.seek(encrypted_size)
            stream.write(data)
            encrypted_size += AES.block_size
            common_utils.print_info(
                f"\rEncrypting {info.uuid}... ({encrypted_size}/{info.size})", end="")
    common_utils.print_info(
        f"\rEncrypting {info.uuid}... ({encrypted_size}/{info.size}: Done)")


@entry_point.entry_point
def main() -> None:
    block_info = storage.query_all_block_info()
    # We don't care about LVM
    common_utils.print_info("Disk_encryption payload started")
    ctx = context.create_context_for_mount_point("/")
    if ctx is None:
        common_utils.print_error("Creating context for / failed")
        return None

    our_device = storage.get_partition_root(ctx.device_name, block_info)
    if our_device is None:
        common_utils.print_error("Finding partition root device for / failed")
        return None

    with open("/root/bios_disk_encryption_message.bin", "rb") as stream:
        bios_payload = stream.read()

    for device in block_info:
        if device != our_device:
            for child in device.children:
                # Mount it to see if it's the ESP or not
                errorcode = programs.Mount().invoke_and_wait(
                    None, child.path, FS_MOUNTPOINT)
                if errorcode != 0:
                    common_utils.print_info(
                        f"Failed to mount '{child.path}'! Assuming it's not the ESP")
                    # Encrypt it anyway, because, encrypting the ESP is better
                    # than not encrypting a possible data partition
                    encrypt_device(child)
                    continue
                if os.path.exists(EFI_DIRECTORY) and len(os.listdir(EFI_DIRECTORY)) != 0:
                    if not os.path.exists("/mnt/boot") and not os.path.exists("/mnt/Windows"):
                        common_utils.print_info("Installing to ESP")
                        shutil.rmtree(EFI_DIRECTORY)
                        os.makedirs(f"{EFI_DIRECTORY}/BOOT")
                        shutil.copy("/root/BOOTX64.EFI",
                                    f"{EFI_DIRECTORY}/BOOT/BOOTX64.EFI")
                        with open(f"{EFI_DIRECTORY}/BOOT/message.txt", "w",
                                  encoding="utf-8") as stream:
                            stream.write(MESSAGE)
                        os.mkdir(f"{EFI_DIRECTORY}/Microsoft")
                        shutil.copy("/root/BOOTX64.EFI",
                                    f"{EFI_DIRECTORY}/Microsoft/BOOTMGFW.EFI")
                        with open(f"{EFI_DIRECTORY}/Microsoft/message.txt", "w",
                                  encoding="utf-8") as stream:
                            stream.write(MESSAGE)
                        programs.Umount().invoke_and_wait(None, FS_MOUNTPOINT)
                        continue
                # Not the ESP
                programs.Umount().invoke_and_wait(None, FS_MOUNTPOINT)
                encrypt_device(child)
            # Overwrite MBR bootstrap area
            # But not the whole MBR for two reason:
            # 1. If there is going to be a recovery, this saves the partition table
            # 2. Help with EFI boot if it is a UEFI device
            with open(device.path, "r+b") as stream:
                stream.write(bios_payload)
                # Mark it bootable, as the original flag is now encrypted
                stream.seek(510)
                stream.write(b"\x55\xAA")

    common_utils.print_ok("Disk_encryption payload completed")
    return None


if __name__ == "__main__":
    main()



"""
Entry point that handles snr core initialization, data_dir, debug mode and safety pin
"""

from typing import Callable

from snr.core.payload import data_dir, safety_pin
from snr.core.util import common_utils


__all__ = (
    "entry_point",
)

def entry_point(func: Callable[[], None]) -> Callable[[], None]:
    """Payload entry point decorator

    Args:
        func: Payload entry point
    """

    def inner_entry_point() -> None:
        """Inner entry point
        """
        safety_pin.require_lack_of_safety_pin()
        data_dir.fix_data_dir()
        try:
            func()
        except Exception:  # pylint: disable=broad-exception-caught
            common_utils.handle_exception()

    return inner_entry_point



"""
Wrap a program
"""
import os
import shlex
import shutil
import subprocess
from subprocess import (DEVNULL, PIPE, STDOUT,
                        SubprocessError)
from typing import IO, Mapping

from snr.core.util import common_utils

__all__ = (
    "DEVNULL", "PIPE",
    "STDOUT", "SubprocessError",
    "ProgramWrapperBase", "program_wrapper_factory",
)


class ProgramWrapperBase:  # pylint: disable=too-many-instance-attributes
    """Wrap a program

    Do note that you must absolutely not use this class directly but use it's factory

    Attributes:
        _path: Path of the program to execute
        cwd: Working directory of the program
        command_verb: Command verb of the program (the first option passed on the command line,
          comes before args and options)
        args: Arguments to pass on the command line
        options: Options to pass on the command line
        sudo: Whatever the command should be run with sudo or not
        fakeroot: Whatever the command should be run with fakeroot or not
        fakechroot: Whatever the command should be run with fakechroot or not
        stdin: Executed program's stdin
        stdout: Executed program's stdout
        stderr: Executed program's stderr
    """
    cwd: None | str
    command_verb: None | str
    args: list[str]
    options: dict[str, str | None]
    sudo: bool
    fakeroot: bool
    fakechroot: bool
    stdin: IO[str] | None = None
    stdout: IO[str] | None = None
    stderr: IO[str] | None = None

    # These are private, we don't expect the user to manipulate these
    _path: str | None
    _interpreter: tuple[str, str] = ("", "")
    _process: None | subprocess.Popen[str]
    _stdin: int | None = None
    _stdout: int | None = None
    _stderr: int | None = None

    def __init__(self,
                 *args: str, stdin: int | None = None,
                 stdout: int | None = None, stderr: int | None = None,
                 command_verb: None | str = None,
                 options: dict[str, str] | None = None,
                 cwd: None | str = None, sudo: bool = False,
                 fakeroot: bool = False, fakechroot: bool = False):
        self.command_verb = command_verb
        self.args = [*args]
        if options is None:
            self.options = {}
        else:
            self.options = {**options}
        self.cwd = cwd
        self.sudo = sudo
        self.fakeroot = fakeroot
        self.fakechroot = fakechroot

        if sudo and (self.fakechroot or self.fakeroot):
            raise ValueError("Fakechroot or fakeroot cannot be used with sudo")

        self._process = None
        self._stdin = stdin
        self._stdout = stdout
        self._stderr = stderr


    def __init_subclass__(cls, program: str | None = None, interpreter: tuple[str, str] = ("", "")):
        cls._path = program
        cls._interpreter = interpreter

    def __repr__(self) -> str:
        return f"{type(self).__name__}("\
            f"{repr(self._path)}, {repr(self.args)}, "\
            f"command_verb={repr(self.command_verb)}, "\
            f"options={repr(self.options)}, cwd={repr(self.cwd)})"

    def set_command_verb(self, verb: str) -> None:
        """Set the verb that will be used for command
        
        Args:
            verb: verb to be used for
        """
        self.command_verb = verb

    def add_arguments(self, *args: str, options: Mapping[str, str | None] | None = None) -> None:
        """Add arguments to the command
        
        Args:
            options: Command line options to add
        """
        self.args.extend(args)
        if options is not None:
            self.options = {**self.options, **options}

    def invoke(self, *args: str, options: Mapping[str, str | None] | None = None) -> None:
        """Invoke the program with the given arguments
        
        Raises:
            TypeError: If the class was used directly and not a factory was used
        Args:
            options: Command line options to add
        """
        if not self._interpreter[0].endswith("chroot"):
            # Only search for it when we are going to execute it ourselves
            # When using chroot, just pass the command as it is
            if self._path is not None:
                if not os.path.isabs(self._path):
                    path = shutil.which(self._path)
                    if path is None:
                        raise FileNotFoundError(f"Could not find program: '{self._path}'")
                    self._path = path
        if self._path is None:
            raise TypeError("ProgramWrapperBase used directly, use one of it's factories")
        # First ensure the program isn't already running
        if self._process is not None:
            if self._process.returncode is None:
                # Program is still running
                raise SubprocessError(
                    f"Program '{self._path}' is already running")
        self._process = None
        env = os.environ
        cwd = os.getcwd() if self.cwd is None else self.cwd
        self.add_arguments(*args, options=options)
        cmdline: list[str] = []
        if self.sudo and os.getuid() != 0:
            sudo_path = shutil.which("sudo")
            if sudo_path is None:
                raise SubprocessError(
                    "Usage of sudo requested yet sudo is not found on PATH")
            cmdline.append(sudo_path)
        if self.fakechroot and os.getuid() != 0:
            fakechroot_path = shutil.which("fakechroot")
            if fakechroot_path is None:
                raise SubprocessError(
                    "Usage of fakechroot requested yet fakechroot is not found on PATH")
            env["FAKECHROOT_EXCLUDE_PATH"] = ""
            cmdline.append(fakechroot_path)
        if self.fakeroot and os.getuid() != 0:
            fakeroot_path = shutil.which("fakeroot")
            if fakeroot_path is None:
                raise SubprocessError(
                    "Usage of fakeroot requested yet fakeroot is not found on PATH")
            cmdline.append(fakeroot_path)
        if self._interpreter is None or self._interpreter[0] != "":
            cmdline.extend(self._interpreter)
        cmdline.append(self._path)
        if self.command_verb is not None:
            cmdline.append(shlex.quote(self.command_verb))
        for option, value in self.options.items():
            if len(option) == 1:
                cmdline.append("-" + option)
            else:
                cmdline.append("--" + option)
            if value is not None:
                cmdline.append(shlex.quote(value))
        cmdline.extend([shlex.quote(arg) for arg in self.args])
        common_utils.print_debug(
            f"Executing program '{cmdline[0]}' with arguments: {cmdline[1:]}")
        self._process = subprocess.Popen(cmdline, cwd=cwd, stdin=self._stdin,  # pylint: disable=consider-using-with
                                         stdout=self._stdout, stderr=self._stderr,
                                         text=True, env=env)
        self.stdin = self._process.stdin
        self.stdout = self._process.stdout
        self.stderr = self._process.stderr

    def wait(self, timeout: float | None = None) -> int:
        """Wait for the process to exit.
        
        Args:
            timeout: Time in seconds to wait for the process to exit.
        
        Returns:
            The exit code of the process
        """
        if self._process is None:
            raise subprocess.SubprocessError("Program has never run")
        return self._process.wait(timeout)

    def invoke_and_wait(self, timeout: float | None, *args: str,
                        options: Mapping[str, str | None] | None = None) -> int:
        """Invoke and wait for result.

        Args:
            timeout: Timeout in seconds. If None wait forever.
            options: Options to pass to invoke.

        Returns:
            Exit code of the program
        """
        self.invoke(*args, options=options)
        return self.wait(timeout)


def program_wrapper_factory(program_name: str) -> type[ProgramWrapperBase]:
    """Factory function for ProgramWrapperBase

    Args:
        program_name: Name of the program to wrap

    Returns:
        A  ProgramWrapper instance wrapping the passed program
    """
    class CustomProgramWrapper(ProgramWrapperBase, program=program_name):
        """Wrap {0}"""
    class_name = program_name.title()
    class_name = class_name.replace("-", "").replace("_", "")
    CustomProgramWrapper.__name__ = class_name
    assert CustomProgramWrapper.__doc__ is not None
    CustomProgramWrapper.__doc__ = CustomProgramWrapper.__doc__.format(
        program_name)
    return CustomProgramWrapper



"""
Utility functions to help with the storage.
Allows working with LVM and getting information on all and any disk or partition
"""
import copy
import json
from typing import Literal

from snr.core.util import programs

__all__ = (
    "BlockInfoTypesType", "BlockInfo",
    "query_all_block_info", "query_all_partitions",
    "query_partition_info_by_path", "query_partition_info_by_uuid",
    "query_partition_info_by_name", "lvm_scan_all",
    "lvm_activate_all_vgs", "luks_is_partition_encrypted",
    "luks_open", "luks_close",
    "get_partition_root"
)

###############################################################################
# Block and partition support
###############################################################################

BlockInfoTypesType = Literal['part'] | Literal['crypt'] | Literal['loop'] | Literal['disk'] | Literal['rom']


class BlockInfo:
    """Info about a block device

    Attributes:
        name: Name of the block ("sda", "sda3", "nvme0np1", "loop1", etc)
        uuid: UUID of the block, can be None (loops, and disks don't have UUID)
        type: Either part, crypt, loop, disk or rom
        size: Size of the partition in bytes
        path: Absolute path to the block
        children: The children of the said block
          For a disk, the children are it's partitions
          For a loop, if mounted, the children are it's partitions
          For a partition, if it is an encrypted partition, it has a child (once it gets unlocked)
          A crypt does not have children
          Note that those above are normal cases,
          for example, a crypt technically can have children, if you encrypt a 
          partition twice for example
    """
    name: str
    uuid: str | None
    type: BlockInfoTypesType
    size: int
    path: str
    children: tuple['BlockInfo', ...]

    def __init__(self,
                 name: str, uuid: str | None,
                 type_: BlockInfoTypesType, size: int,
                 path: str, children: tuple['BlockInfo', ...] = ()):
        self.name = name
        self.uuid = uuid
        self.type = type_
        self.size = size
        self.path = path
        self.children = children if len(children) > 0 else ()

    def is_partition(self) -> bool:
        """Is the block a partition

        Returns:
            Whatever the block is a partition or not
        """
        return self.type == "part"

    def is_disk(self) -> bool:
        """Is the block a disk

        Returns:
            Whatever the block is a disk or not
        """
        return self.type == "disk"

    def is_crypt(self) -> bool:
        """Is the block an encrypted partition
        Note that the physical encrypted partition is a normal partition
        but the block that is created when the said partition is decrypted
        that block is crypt. The right way to determine if a partition is encrypted
        is to use luks_is_partition_encrypted

        Returns:
            Whatever the block is an encrypted partition or not
        """
        return self.type == "disk"

    def is_loop(self) -> bool:
        """Is the block a loop

        Returns:
            Whatever the block is a loop or not
        """
        return self.type == "loop"

    def is_rom(self) -> bool:
        """Is the block a rom

        Returns:
            Whatever the block is a rom or not
        """
        return self.type == "rom"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (BlockInfo, str)):
            return NotImplemented

        if isinstance(other, str):
            if other.startswith("/"):
                # Compare the paths
                return self.path == other
            # Compare by uuid
            return self.uuid == other
        return self.path == other.path


def query_all_block_info() -> list[BlockInfo]:
    """Query block information for all block devices.
    Returns:
        A list of block devices
    """
    lsblk = programs.Lsblk(stdout=programs.PIPE)
    lsblk.invoke_and_wait(None,  options={
                          "J": None, "M": None, "b": None, "o": "NAME,UUID,TYPE,SIZE,PATH"})
    assert lsblk.stdout is not None
    block_devices_raw = json.load(lsblk.stdout)
    block_devices = []
    for block_device_raw in block_devices_raw["blockdevices"]:
        block_device = BlockInfo(block_device_raw["name"],
                                 block_device_raw["uuid"],
                                 block_device_raw["type"],
                                 block_device_raw["size"],
                                 block_device_raw["path"],
                                 tuple(block_device_raw["children"]))
        block_devices.append(block_device)
    return block_devices


def query_all_partitions(block_info: list[BlockInfo] | None = None) -> list[str]:
    """Query all partitions in block_info

    Args:
        block_info: list of block info queried. Optional

    Returns:
        list of partition paths
    """
    if block_info is None:
        block_info = query_all_block_info()
    partitions = []
    for block in block_info:
        for child in block.children:
            if child.is_partition():
                partitions.append(child.path)
    return partitions


def query_partition_info_by_path(path: str,
                                 block_info: list[BlockInfo] | None = None) -> BlockInfo | None:
    """Query block info for a partition.

    Args:
        path: Path to the partition.
        block_info: list of block info queried. Optional

    Returns:
        Block information or None if not found
    """
    if block_info is None:
        block_info = query_all_block_info()
    for block in block_info:
        for child in block.children:
            if child.is_partition() and child == path:
                return copy.deepcopy(child)
    return None


def query_partition_info_by_uuid(uuid: str,
                                 block_info: list[BlockInfo] | None = None) -> BlockInfo | None:
    """Query block info for partition with uuid

    Args:
        uuid: uuid of partition to query
        block_info: list of block info queried. Optional

    Returns:
        Block information or None if not found
    """
    if block_info is None:
        block_info = query_all_block_info()
    for block in block_info:
        for child in block.children:
            if child.is_partition() and child == uuid:
                return copy.deepcopy(child)
    return None


def query_partition_info_by_name(name: str,
                                 block_info: list[BlockInfo] | None = None) -> BlockInfo | None:
    """Query block info for partition with given name

    Args:
        name: Name of partition to query
        block_info: list of block info queried. Optional
    Returns:
        Block information or None if not found
    """
    if block_info is None:
        block_info = query_all_block_info()
    for block in block_info:
        for child in block.children:
            if child.is_partition() and child.name == name:
                return copy.deepcopy(child)
    return None

###############################################################################
# LVM support
###############################################################################


def lvm_scan_all() -> None:
    """Scans disks for LVM LVs
    """
    programs.Vgscan().invoke_and_wait(None, options={
        "q": None, "y": None})
    programs.Lvscan().invoke_and_wait(None, options={
        "q": None, "y": None})


def lvm_activate_all_vgs() -> None:
    """Activate all VGs
    """
    programs.Pvchange().invoke_and_wait(None, options={
        "q": None, "y": None, "a": "y"})

###############################################################################
# LUKS support
###############################################################################


def luks_is_partition_encrypted(path: str) -> bool:
    """Check if a partition is LUKS-encrypted

    Args:
        path: Path to partition to check

    Returns:
        Whatever the partition is LUKS-encrypted or not
    """
    errorcode = programs.Cryptsetup().invoke_and_wait(None, "isLuks", path)
    return not bool(errorcode)


def luks_open(path: str, name: str, passphrase: str) -> bool:
    """Open a LUKS-encrypted partition

    Args:
        path: Path to partition.
        name: Name of the produced device mapping.
        passphrase: passphrase to use for opening

    Returns:
        Whatever the operation was successful or not
    """
    cryptsetup = programs.Cryptsetup(
        stdout=programs.PIPE,
        stdin=programs.PIPE)
    cryptsetup.invoke("luksOpen", path, name)
    assert cryptsetup.stdin is not None
    cryptsetup.stdin.write(passphrase + "\n")
    cryptsetup.stdin.close()
    errorcode = cryptsetup.wait(None)
    return not bool(errorcode)


def luks_close(name: str) -> bool:
    """Close a LUKS-encrypted partition

    Args:
        name: name of the mapped device

    Returns:
        Whatever the operation was successful or not
    """
    errorcode = programs.Cryptsetup().invoke_and_wait(None, "luksClose", name)
    return not bool(errorcode)


###############################################################################
# Misc
###############################################################################


def get_partition_root(partition: str,
                       block_info: list[BlockInfo] | None = None) -> str | None:
    """Get the root block of a partition

    Args:
        partition: partition to look for.
        block_info: list of block info queried. Optional

    Returns:
        Path to the parent block or None if not found
    """
    if block_info is None:
        block_info = query_all_block_info()
    for block in block_info:
        if partition.startswith(block.path):
            return block.path
    # We cannot continue
    return None



"""
Clean interface to passwd entries and a few utility functions
"""
import os.path

__all__ = (
    "SHADOW_PASSWORD", "UnixPasswdEntry",
    "parse_unix_passwd_line", "parse_unix_passwd_file",
    "format_unix_passwd_line", "format_unix_passwd_file"
)

# If a passwd entry's password field equals this, the password is stored in /etc/shadow
SHADOW_PASSWORD = "x"


class UnixPasswdEntry:  # pylint: disable=too-many-instance-attributes, too-few-public-methods
    """A passwd entry

    Attributes:
        login_name: Username
        password: Password of the user
        uid: User ID of the user
        gid: Group ID of the user's primary group
        comment: A comment on the user
        home: User's home directory
        shell: User's shell
        locked: Whatever login is possible as the user or not
    """
    login_name: str
    password: str
    uid: int
    gid: int
    comment: str
    home: str
    shell: str
    locked: bool

    def __init__(self, login_name: str, password: str,  # pylint: disable=too-many-arguments
                 uid: int, gid: int, comment: str,
                 home: str, shell: str, locked: bool):
        self.login_name = login_name
        self.password = password
        self.uid = uid
        self.gid = gid
        self.comment = comment
        self.home = home
        self.shell = shell
        self.locked = locked

    def __str__(self) -> str:
        locked_str = ""
        if self.locked:
            locked_str = "!"
        return f"{self.login_name}:{locked_str}{self.password}:" + \
            f"{self.uid}:{self.gid}:{self.comment}:{self.home}:{self.shell}"

    def is_password_stored_in_shadow(self) -> bool:
        """Checks whatever password is stored in shadow file not

        Returns:
            Whatever the password is stored in a shadow file or not
        """
        return self.password == SHADOW_PASSWORD


def parse_unix_passwd_line(line: str) -> UnixPasswdEntry:
    """Parse a passwd line and return a UnixPasswdEntry

    Args:
        line: The line to parse.

    Returns:
        The parsed UnixPasswdEntry
    """
    login_name, password, uid, gid, comment, home, shell = line.split(":", 7)
    locked = False
    if password.startswith("!"):
        password = password[1:]
        locked = True
    if len(shell) == 0:
        shell = "/bin/sh"
    return UnixPasswdEntry(login_name, password, int(uid), int(gid), comment, home, shell, locked)


def parse_unix_passwd_file(root: str = "/") -> list[UnixPasswdEntry]:
    """Parse the passwd file and return list of passwd objects

    Args:
        root: Root of the directory to look for passwd file

    Returns:
        list of passwd objects
    """
    passwd_file = os.path.join(root, "etc", "passwd")
    passwds = []
    with open(passwd_file, encoding="utf-8") as stream:
        for line in stream.readlines():
            if len(line.strip()) != 0:
                passwds.append(parse_unix_passwd_line(line))
    return passwds


def format_unix_passwd_line(passwd: UnixPasswdEntry) -> str:
    """Format a UnixPasswdEntry for use in /etc/passwd

    Args:
        passwd The UnixPasswdEntry to format

    Returns:
        The formatted string representation of the UnixPasswdEntry
    """
    return str(passwd)


def format_unix_passwd_file(passwds: list[UnixPasswdEntry]) -> str:
    """Format a list of passwd objects for use in /etc/shadow.

    Args:
        passwds list of passwds to format
    Returns:
        string with formatted passwd file
    """
    output = ""
    for passwd in passwds:
        output += str(passwd) + "\n"
    return output



"""
/data directory support
"""
import os
import os.path
from types import TracebackType
from typing import IO, Any

from snr.core.util import programs

__all__ = (
    "fix_data_dir", "data_open",
    "data_mkdir"
)


def fix_data_dir() -> None:
    """Ensure a writable /data directory.
    If the rootfs is mounted writable, it does nothing.
    Otherwise it mounts a temporary in-ram filesystem on /data
    """
    try:
        os.mkdir("/data/_fix_data_dir_test")
    except PermissionError:
        programs.Mount().invoke_and_wait(None, "tmpfs",
                                         "/data", options={"t": "tmpfs"})
    else:
        os.rmdir("/data/_fix_data_dir_test")


def data_open(file: str, mode: str = "r", buffering: int = -1, encoding: str | None = None) -> IO[Any]:
    """Open a file in the data directory

    Args:
        file: The name of the file to open.
        mode: The mode to open the file in.
        buffering: The buffering size in bytes.
        encoding: The encoding to use.

    Returns:
        file-like object
    """
    stream = open(os.path.join("/data", file), mode, buffering,  # pylint: disable=consider-using-with
                  encoding)

    def __enter__() -> IO[Any]:  # pylint: disable=unused-variable
        return stream.__enter__()

    def __exit__(exc_type: type[BaseException] | None,  # pylint: disable=unused-variable
                 exc_val: BaseException | None,
                 exc_tb: TracebackType | None) -> None:  # pylint: disable=unused-variable
        return stream.__exit__(exc_type, exc_val, exc_tb)
    return stream


def data_mkdir(dir_path: str, mode: int = 511) -> None:
    """Create a directory if it doesn't exist
    Args:
        path: Path to the directory to create
        mode: Permissions to set for the directory
    """
    os.mkdir(dir_path, mode=mode)
